import inspect
import os
from functools import wraps

import mock
import pytest

from doctor.errors import (
    ForbiddenError, ImmutableError, InvalidValueError, NotFoundError,
    UnauthorizedError)
from doctor.flask import (
    handle_http, HTTP400Exception, HTTP401Exception, HTTP403Exception,
    HTTP404Exception, HTTP409Exception, HTTP500Exception,
    should_raise_response_validation_errors)
from doctor.response import Response
from doctor.utils import (
    add_param_annotations, get_params_from_func, Params, RequestParamAnnotation)

from .types import (
    Auth, Colors, FooInstance, Item, ItemId, IncludeDeleted, Latitude)
from .utils import add_doctor_attrs


def check_auth(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)
    params = [
        RequestParamAnnotation('auth', Auth, True),
    ]
    _wrapper = add_param_annotations(wrapper, params)
    return _wrapper


def get_item(item_id: ItemId, include_deleted: IncludeDeleted = False) -> Item:
    return {'item_id': 1}


def create_item(item: Item, colors: Colors, optional_id: ItemId = None,
                lat: Latitude = None) -> Item:
    return {'item_id': 1}


@pytest.fixture
def mock_request():
    mock_request_patch = mock.patch('doctor.flask.request')
    yield mock_request_patch.start()
    mock_request_patch.stop()


@pytest.fixture
def mock_get_logic():
    mock_logic = mock.MagicMock(spec=get_item, return_value={'item_id': 1})
    mock_logic._doctor_signature = inspect.signature(get_item)
    mock_logic._doctor_params = get_params_from_func(mock_logic)
    mock_logic._doctor_allowed_exceptions = None
    mock_logic._doctor_req_obj_type = None
    return mock_logic


@pytest.fixture
def mock_post_logic():
    mock_logic = mock.MagicMock(spec=create_item, return_value={'item_id': 1})
    mock_logic._doctor_signature = inspect.signature(create_item)
    mock_logic._doctor_params = get_params_from_func(mock_logic)
    mock_logic._doctor_allowed_exceptions = None
    mock_logic._doctor_req_obj_type = None
    return mock_logic


def test_handle_http_with_json(mock_request, mock_post_logic):
    mock_request.method = 'POST'
    mock_request.content_type = 'application/json; charset=UTF8'
    mock_request.mimetype = 'application/json'
    mock_request.json = {
        'item': {
            'item_id': 1,
        },
        'colors': ['blue'],
        'optional_id': None,
        'location.lat': 45.2342343,
    }
    mock_handler = mock.Mock()

    actual = handle_http(mock_handler, (), {}, mock_post_logic)
    assert actual == ({'item_id': 1}, 201)

    expected_call = mock.call(
        item={'item_id': 1}, colors=['blue'], optional_id=None,
        lat=45.2342343)
    assert expected_call == mock_post_logic.call_args


def test_handle_http_with_route_that_defines_req_obj_type(mock_request):
    def logic(foo: FooInstance):
        return foo

    logic = add_doctor_attrs(logic, req_obj_type=FooInstance)

    mock_request.method = 'POST'
    mock_request.content_type = 'application/json; charset=UTF8'
    mock_request.mimetype = 'application/json'
    mock_request.json = {
        'foo': 'A foo',
        'foo_id': 1,
    }
    mock_handler = mock.Mock()
    actual = handle_http(mock_handler, (), {}, logic)
    assert actual == ({'foo_id': 1, 'foo': 'A foo'}, 201)


def test_handle_http_object_array_types(mock_request, mock_post_logic):
    """
    This test verifies that we pass native types to the logic function rather
    than the doctor types that annotate the parameters of the logic function.
    This is done to prevent downstream issues from happening by passing say
    an Integer instance instead of an int to pymysql.  pymysql doesn't know
    about this type and tries to escape it like a string which causes an
    AttributeError.
    """
    mock_request.method = 'POST'
    mock_request.content_type = 'application/json; charset=UTF8'
    mock_request.mimetype = 'application/json'
    mock_request.json = {'item': {'item_id': 1}, 'colors': ['blue', 'green']}

    mock_handler = mock.Mock()
    actual = handle_http(mock_handler, (), {}, mock_post_logic)
    assert actual == ({'item_id': 1}, 201)

    expected_call = mock.call(item={'item_id': 1}, colors=['blue', 'green'])
    assert expected_call == mock_post_logic.call_args

    # Verify we're passing native types and not doctor types to logic functions
    kwargs = mock_post_logic.call_args[1]
    assert type(kwargs['item']) is dict
    assert type(kwargs['colors']) is list


def test_handle_http_non_json(mock_request, mock_get_logic):
    mock_request.method = 'GET'
    mock_request.content_type = 'application/x-www-form-urlencoded'
    mock_request.values = {'item_id': '3'}
    mock_handler = mock.Mock()

    actual = handle_http(mock_handler, (), {}, mock_get_logic)
    assert actual == ({'item_id': 1}, 200)

    expected_call = mock.call(item_id=3)
    assert expected_call == mock_get_logic.call_args


def test_handle_http_unsupported_http_method_with_body(
        mock_request, mock_get_logic):
    mock_request.method = 'GET'
    mock_request.content_type = 'application/json; charset=UTF8'
    mock_request.mimetype = 'application/json'
    mock_request.values = {'item_id': '3'}
    mock_handler = mock.Mock()

    actual = handle_http(mock_handler, (), {}, mock_get_logic)
    assert actual == ({'item_id': 1}, 200)

    expected_call = mock.call(item_id=3)
    assert expected_call == mock_get_logic.call_args


def test_handle_http_missing_required_arg(mock_request, mock_get_logic):
    mock_request.method = 'GET'
    mock_request.content_type = 'application/x-www-form-urlencoded'
    mock_request.values = {}
    mock_handler = mock.Mock()

    with pytest.raises(HTTP400Exception, match='item_id is required'):
        handle_http(mock_handler, (), {}, mock_get_logic)


def test_handle_http_missing_multiple_required_args(
        mock_request, mock_post_logic):
    mock_request.method = 'POST'
    mock_request.content_type = 'application/json; charset=UTF8'
    mock_request.mimetype = 'application/json'
    mock_request.json = {}

    mock_handler = mock.Mock()
    with pytest.raises(HTTP400Exception,
                       match="\['item', 'colors'\] are required."):
        handle_http(mock_handler, (), {}, mock_post_logic)


def test_handle_http_multiple_invalid_args(mock_request, mock_post_logic):
    mock_request.method = 'POST'
    mock_request.content_type = 'application/json; charset=UTF8'
    mock_request.mimetype = 'application/json'
    mock_request.json = {'item': 1, 'colors': 'blue'}

    mock_handler = mock.Mock()
    expected_msg = ("{'item': 'Must be an object.', "
                    "'colors': 'Must be a list.'}")
    with pytest.raises(HTTP400Exception, match=expected_msg):
        handle_http(mock_handler, (), {}, mock_post_logic)


def test_handle_http_decorator_adds_param_annotations(
        mock_request, mock_get_logic):
    """
    This test verifies if a decorator uses doctor.utils.add_param_annotations
    to add params to the logic function that we fail to validate if the added
    params are missing or invalid.
    """
    mock_request.method = 'GET'
    mock_request.content_type = 'application/x-www-form-urlencoded'
    mock_request.values = {'item_id': '1'}
    mock_handler = mock.Mock()
    logic = check_auth(add_doctor_attrs(get_item))

    expected_params = Params(all=['item_id', 'include_deleted', 'auth'],
                             optional=['include_deleted'],
                             required=['item_id', 'auth'],
                             logic=['item_id', 'include_deleted'])
    assert expected_params == logic._doctor_params
    with pytest.raises(HTTP400Exception, match='auth is required'):
        handle_http(mock_handler, (), {}, logic)

    # Add auth and it should validate
    mock_request.values = {'item_id': '1', 'auth': 'auth'}
    actual = handle_http(mock_handler, (), {}, logic)
    assert actual == ({'item_id': 1}, 200)


def test_handle_http_invalid_param(mock_request, mock_get_logic):
    mock_request.method = 'GET'
    mock_request.content_type = 'application/x-www-form-urlencoded'
    mock_request.values = {'item_id': 'string'}
    mock_handler = mock.Mock()

    expected_msg = 'item_id - value must be a valid type \(integer, null\)'
    with pytest.raises(HTTP400Exception, match=expected_msg):
        handle_http(mock_handler, (), {}, mock_get_logic)


@mock.patch('doctor.flask.current_app')
def test_handle_http_allowed_exception(mock_app, mock_request, mock_get_logic):
    mock_app.config = {'DEBUG': False}
    mock_request.method = 'GET'
    mock_request.content_type = 'application/x-www-form-urlencoded'
    mock_request.values = {'item_id': '1'}
    mock_handler = mock.Mock()
    mock_get_logic.side_effect = ValueError('Allowed')
    mock_get_logic._doctor_allowed_exceptions = [ValueError]

    with pytest.raises(ValueError, match='Allowed'):
        handle_http(mock_handler, (), {}, mock_get_logic)


def test_handle_http_response_instance_return_value(
        mock_request, mock_get_logic):
    mock_request.method = 'GET'
    mock_request.content_type = 'application/json; charset=UTF8'
    mock_request.mimetype = 'application/json'
    mock_request.values = {'item_id': '3'}
    mock_get_logic.return_value = Response({'item_id': 3}, {'X-Header': 'Foo'},
                                           status_code=202)
    mock_handler = mock.Mock()

    actual = handle_http(mock_handler, (), {}, mock_get_logic)
    assert actual == ({'item_id': 3}, 202, {'X-Header': 'Foo'})


def test_should_raise_response_validation_errors():
    assert should_raise_response_validation_errors() is False

    os.environ['RAISE_RESPONSE_VALIDATION_ERRORS'] = '1'
    assert should_raise_response_validation_errors() is True


@mock.patch('doctor.flask.current_app')
def test_handle_http_response_validation(
        mock_app, mock_request, mock_get_logic):
    mock_app.config = {'DEBUG': False}

    mock_request.method = 'GET'
    mock_request.content_type = 'application/x-www-form-urlencoded'
    mock_request.values = {'item_id': '3'}
    mock_handler = mock.Mock()
    mock_get_logic.return_value = {'foo': 'bar'}

    expected = ("{'item_id': 'This field is required.', "
                "'foo': 'Additional properties are not allowed.'}")
    with pytest.raises(HTTP400Exception, match=expected):
        handle_http(mock_handler, (), {}, mock_get_logic)

    # Should also work with the response is an instance of Response
    mock_get_logic.return_value = Response({'foo': 'bar'})
    with pytest.raises(HTTP400Exception, match=expected):
        handle_http(mock_handler, (), {}, mock_get_logic)


@mock.patch('doctor.flask.should_raise_response_validation_errors')
@mock.patch('doctor.flask.current_app')
def test_handle_http_response_type_validation(
        mock_app, mock_should, mock_request):
    """
    This tests that if we provide a type to our response that it validates
    the content of the response against that type.
    """
    def get_logic() -> Response[Item]:
        return Response({'foo': 'bar'})

    get_logic = add_doctor_attrs(get_logic)
    mock_should.return_value = True
    mock_app.config = {'DEBUG': False}

    mock_request.method = 'GET'
    mock_request.content_type = 'application/x-www-form-urlencoded'
    mock_request.values = {'item_id': '3'}
    mock_handler = mock.Mock()

    expected = ("{'item_id': 'This field is required.', "
                "'foo': 'Additional properties are not allowed.'}")
    with pytest.raises(HTTP400Exception, match=expected):
        handle_http(mock_handler, (), {}, get_logic)

    # Change the return value and validate it passes response validation.
    def get_logic() -> Response[Item]:
        return Response({'item_id': 1})

    get_logic = add_doctor_attrs(get_logic)
    actual = handle_http(mock_handler, (), {}, get_logic)
    assert ({'item_id': 1}, 200, None) == actual


@mock.patch('doctor.flask.should_raise_response_validation_errors')
@mock.patch('doctor.flask.current_app')
def test_handle_http_http_errors(
        mock_app, mock_should, mock_request, mock_get_logic):
    mock_app.config = {'DEBUG': False}

    mock_request.method = 'GET'
    mock_request.content_type = 'application/x-www-form-urlencoded'
    mock_request.values = {'item_id': '3'}
    mock_handler = mock.Mock()

    # 400
    mock_get_logic.side_effect = InvalidValueError('invalid')
    with pytest.raises(HTTP400Exception, match='invalid'):
        handle_http(mock_handler, (), {}, mock_get_logic)

    # 401
    mock_get_logic.side_effect = UnauthorizedError('unauth')
    with pytest.raises(HTTP401Exception, match='unauth'):
        handle_http(mock_handler, (), {}, mock_get_logic)

    # 403
    mock_get_logic.side_effect = ForbiddenError('forbidden')
    with pytest.raises(HTTP403Exception, match='forbidden'):
        handle_http(mock_handler, (), {}, mock_get_logic)

    # 404
    mock_get_logic.side_effect = NotFoundError('not found')
    with pytest.raises(HTTP404Exception, match='not found'):
        handle_http(mock_handler, (), {}, mock_get_logic)

    # 409
    mock_get_logic.side_effect = ImmutableError('immutable')
    with pytest.raises(HTTP409Exception, match='immutable'):
        handle_http(mock_handler, (), {}, mock_get_logic)

    # 500
    mock_get_logic.side_effect = Exception('internal error')
    with pytest.raises(HTTP500Exception, match='Uncaught error in logic func'):
        handle_http(mock_handler, (), {}, mock_get_logic)

    # 500 in debug mode
    mock_app.config = {'DEBUG': True}
    with pytest.raises(Exception, match='internal error'):
        handle_http(mock_handler, (), {}, mock_get_logic)



import os
from datetime import date, datetime

import pytest

from doctor.errors import TypeSystemError
from doctor.resource import ResourceSchema
from doctor.types import (
    array, boolean, enum, integer, json_schema_type, new_type, number, string,
    Object, MissingDescriptionError, SuperType)


class TestSuperType(object):

    def test_init_no_description(self):
        class MyType(SuperType):
            pass

        with pytest.raises(ValueError, match='define a description attribute'):
            MyType()

    def test_init_with_description_defined(self):
        class MyType(SuperType):
            description = 'My Type'

        MyType()


class TestString(object):

    def test_type(self):
        S = string('string')
        assert type(S('string')) is str

    def test_type_nullable(self):
        S = string('string', nullable=True)
        assert S(None) is None

    def test_trim_whitespace(self):
        S = string('a string', trim_whitespace=True)
        actual = S(' foo ')
        assert 'foo' == actual

    def test_min_length(self):
        S = string('a string', min_length=2)
        # No exception
        S('foo')
        # String too short
        with pytest.raises(TypeSystemError,
                           match='Must have at least 2 characters'):
            S('f')
        # Empty string when min_length is 1
        S = string('empty', min_length=1)
        with pytest.raises(TypeSystemError, match='Must not be blank.'):
            S('')

    def test_max_length(self):
        S = string('a string', max_length=2)
        # No exception
        S('12')
        # String too long
        with pytest.raises(TypeSystemError,
                           match='Must have no more than 2 characters'):
            S('123')

    def test_pattern(self):
        S = string('a regex', pattern=r'^foo*')
        # No exception
        S('foo bar')
        # Does not begin with `foo`
        with pytest.raises(TypeSystemError):
            S('bar')

    def test_format_date(self):
        S = string('date', format='date')
        # No exception
        s = S('2018-10-22')
        assert s == date(2018, 10, 22)
        # Invalid date
        expected_msg = "time data 'foo' does not match format '%Y-%m-%d'"
        with pytest.raises(TypeSystemError, match=expected_msg):
            S('foo')

    def test_format_date_time(self):
        S = string('date-time', format='date-time')
        # No exception
        s = S('2018-10-22T11:12:00')
        assert s == datetime(2018, 10, 22, 11, 12, 0)

        # Invalid datetime
        expected_msg = ("ISO 8601 time designator 'T' missing. Unable to parse "
                        "datetime string 'foo'")
        with pytest.raises(TypeSystemError, match=expected_msg):
            S('foo')

    def test_format_email(self):
        S = string('email', format='email')
        # No exception
        S('user@domain.net')
        # Invalid email
        with pytest.raises(TypeSystemError, match='Not a valid email address.'):
            S('foo.net')

    def test_format_time(self):
        S = string('time', format='time')
        # no exception
        s = S('13:10:00')
        assert 13 == s.hour
        assert 10 == s.minute
        assert 0 == s.second

        # invalid
        expected_msg = "time data 'foo' does not match format '%H:%M:%S"
        with pytest.raises(TypeSystemError, match=expected_msg):
            S('foo')

    def test_format_uri(self):
        S = string('uri', format='uri')
        # no exception
        S('https://doctor.com')
        # Invalid uri
        expected_msg = "'foo' is not a valid 'URI'."
        with pytest.raises(TypeSystemError, match=expected_msg):
            S('foo')

    def test_get_example(self):
        S = string('A string.', example='Foo')
        assert 'Foo' == S.get_example()

        # Default example
        S = string('A string.')
        assert 'string' == S.get_example()


class TestNumericType(object):

    def test_nullable(self):
        N = number('int', nullable=True)
        assert N(None) is None

    def test_minimum(self):
        N = number('float', minimum=3.22)
        # No exception
        N(3.23)
        # Number too small
        with pytest.raises(TypeSystemError,
                           match='Must be greater than or equal to 3.22'):
            N(3.21)

    def test_maximum(self):
        N = number('int', maximum=30)
        # No exception
        N(30)
        # Number > max
        with pytest.raises(TypeSystemError,
                           match='Must be less than or equal to 30'):
            N(31)

    def test_exclusive_mimimum(self):
        N = integer('int', minimum=2, exclusive_minimum=True)
        # No exception
        N(3)
        # number is equal to minimum
        with pytest.raises(TypeSystemError,
                           match='Must be greater than 2.'):
            N(2)

    def test_exclusive_maximum(self):
        N = number('float', maximum=3.3, exclusive_maximum=True)
        # No exception
        N(3.2)
        # number is equal to maximum
        with pytest.raises(TypeSystemError, match='Must be less than 3.3.'):
            N(3.3)

    def test_multiple_of(self):
        N = integer('int', multiple_of=2)
        # No exception
        N(4)
        # Not multiple
        with pytest.raises(TypeSystemError, match='Must be a multiple of 2.'):
            N(5)

        # Also test with float
        N = number('float', multiple_of=1.1)
        # No exception
        N(2.2)
        # Not multiple
        with pytest.raises(TypeSystemError, match='Must be a multiple of 1.1.'):
            N(5)


class TestNumber(object):

    def test_type(self):
        N = number('float')
        assert type(N(3.14)) is float

    def test_get_example(self):
        N = number('A float', example=1.12)
        assert 1.12 == N.get_example()

        # Default example
        N = number('Pi')
        assert 3.14 == N.get_example()


class TestInteger(object):

    def test_type(self):
        I = integer('int')  # noqa
        assert type(I(1)) is int

    def test_get_example(self):
        I = integer('An int', example=1022)  # noqa
        assert 1022 == I.get_example()

        # Default example
        I = integer('An ID')  # noqa
        assert 1 == I.get_example()


class TestBoolean(object):

    def test_type(self):
        B = boolean('bool')
        assert type(B('true')) is bool

    def test_nullable(self):
        B = boolean('bool', nullable=True)
        assert B(None) is None

    def test_boolean_type(self):
        B = boolean('A bool')
        tests = (
            # (input, expected)
            ('true', True),
            ('false', False),
            ('True', True),
            ('False', False),
            ('on', True),
            ('off', False),
            ('1', True),
            ('0', False),
            ('', False),
        )
        for val, expected in tests:
            assert expected == B(val)

    def test_non_boolean(self):
        B = boolean('bool')
        with pytest.raises(TypeSystemError, match='Must be a valid boolean.'):
            B('dog')

    def test_get_example(self):
        B = boolean('A bool', example=False)
        assert B.get_example() is False

        # Default example
        B = boolean('A bool')
        assert B.get_example() is True


class TestEnum(object):

    def test_type(self):
        E = enum('choices', enum=['foo'])
        assert type(E('foo')) is str

    def test_nullalbe(self):
        E = enum('choices', enum=['foo'], nullable=True)
        assert E(None) is None

    def test_enum(self):
        E = enum('choices', enum=['foo', 'bar'])
        # no exception
        E('foo')
        E('bar')
        # not in choices
        expected_msg = "Must be one of: \['foo', 'bar'\]"
        with pytest.raises(TypeSystemError, match=expected_msg):
            E('dog')

    def test_get_example(self):
        E = enum('choices', enum=['foo', 'bar'], example='bar')
        assert 'bar' == E.get_example()

        # Default example (1st item in enum)
        E = enum('choices', enum=['foo', 'bar'])
        assert 'foo' == E.get_example()


class FooObject(Object):
    additional_properties = True
    description = 'A Foo'
    properties = {'foo': string('foo property', min_length=2)}


class NullableFooObject(FooObject):
    nullable = True


class NoAddtPropsObject(FooObject):
    additional_properties = False


class RequiredPropsObject(Object):
    description = 'required'
    properties = {
        'foo': string('foo property', min_length=2),
        'bar': integer('an int'),
    }
    required = ['bar']


class TestObject(object):

    def test_valid_object(self):
        expected = {'foo': 'bar'}
        actual = FooObject(expected)
        assert expected == actual

    def test_nullable(self):
        # Just test that this does not raise an exception.
        NullableFooObject(None)

    def test_missing_description(self):
        class MyObject(Object):
            pass

        expected_msg = "MyObject'> did not define a description attribute"
        with pytest.raises(MissingDescriptionError, match=expected_msg):
            MyObject({'foo': 'bar'})

    def test_invalid_object(self):
        with pytest.raises(TypeSystemError, match='Must be an object.'):
            FooObject('12')

    def test_invalid_property(self):
        with pytest.raises(TypeSystemError,
                           match="{'foo': 'Must have at least 2 characters.'}"):
            FooObject({'foo': 'f'})

    def test_additional_properties(self):
        expected = {
            'foo': 'bar',
            'cat': 12  # Extra property not defined for FooObject
        }
        actual = FooObject(expected)
        assert expected == actual

        expected_msg = "{'cat': 'Additional properties are not allowed.'}"
        with pytest.raises(TypeSystemError, match=expected_msg):
            NoAddtPropsObject(expected)

    def test_required_properties(self):
        expected = {'bar': 1}
        actual = RequiredPropsObject(expected)
        assert expected == actual

        # omit required property 'bar'
        expected = {'foo': 'bar'}
        with pytest.raises(TypeSystemError,
                           match="{'bar': 'This field is required.'}"):
            RequiredPropsObject(expected)

    def test_get_example(self):
        # default example is generated from example values of it's properties.
        assert {'foo': 'string', 'bar': 1} == RequiredPropsObject.get_example()
        # with a defined example
        setattr(RequiredPropsObject, 'example', {'foo': 'foo', 'bar': 33})
        assert {'foo': 'foo', 'bar': 33} == RequiredPropsObject.get_example()


class TestArray(object):

    def test_nullable(self):
        A = array('array', items=string('string', max_length=1), nullable=True)
        # Just tests that this does not raise an exception.
        A(None)

    def test_items(self):
        A = array('array', items=string('string', max_length=1))
        # no exception
        A(['a', 'b'])
        # Invalid type of items
        with pytest.raises(TypeSystemError,
                           match="{0: 'Must have no more than 1 characters.'}"):
            A(['aa', 'b'])

    def test_items_multiple_types(self):
        A = array('a', items=[
            string('string', max_length=1), integer('int', maximum=1234)])
        # no exception
        A(['b', 1234])
        # Invalid type
        with pytest.raises(TypeSystemError,
                           match="{1: 'Must be less than or equal to 1234.'}"):
            A(['b', 1235])

    def test_additional_items(self):
        A = array('a', items=[
            string('string', max_length=1), integer('int', maximum=1234)],
            additional_items=True)
        # no exception
        A(['a', 2, 3, 4, True])

        A = array('a', items=[
            string('string', max_length=1), integer('int', maximum=1234)],
            additional_items=False)
        with pytest.raises(TypeSystemError, match='Too many items.'):
            A(['a', 2, 3, 4, True])

    def test_min_items(self):
        A = array('a', min_items=3)
        # no exception
        A([1, 2, 3])
        # Too few items
        with pytest.raises(TypeSystemError, match='Not enough items'):
            A([1])

    def test_max_items(self):
        A = array('a', max_items=3)
        # no exception
        A([1, 2, 3])
        # Too many items
        with pytest.raises(TypeSystemError, match='Too many items'):
            A([1, 2, 3, 4])

    def test_unique_items(self):
        A = array('not unique', unique_items=False)
        # no exception
        A([1, 1, 1, 2])

        A = array('unique', unique_items=True)
        with pytest.raises(TypeSystemError, match='This item is not unique.'):
            A([1, 1, 1, 2])

    def test_get_example(self):
        A = array('No example of items')
        assert [1] == A.get_example()

        A = array('Defined example', example=['a', 'b'])
        assert ['a', 'b'] == A.get_example()

        A = array('No example, defined items', items=string('letter'))
        assert ['string'] == A.get_example()


class TestJsonSchema(object):

    def test_no_definition_key(self):
        schema_file = os.path.join(
            os.path.dirname(__file__), 'schema', 'annotation.yaml')
        J = json_schema_type(schema_file=schema_file)

        # Verify description and example were set as attributes based
        # on values from the loaded json schema.
        assert 'An annotation object.' == J.description
        assert {
            'annotation_id': 1,
            'auth': 'token',
            'more_id': 1,
            'name': 'Annotation',
            'url': 'https://upsight.com',
        } == J.example
        assert isinstance(J.schema, ResourceSchema)

        # no exception
        data = {
            'annotation_id': 1,
            'name': 'test',
        }
        actual = J(data)
        assert data == actual
        # Verify description pulled from yaml file
        expected = 'An annotation object.'
        assert expected == J.description

        # missing required field
        data = {'name': 'test'}
        with pytest.raises(TypeSystemError,
                           match="'annotation_id' is a required property"):
            J(data)

        # annotation_id is not an int.
        data = {
            'auth': 'authtoken',  # verifying we can load from external files
            'annotation_id': 'foobar',
            'name': 'test',
        }
        with pytest.raises(TypeSystemError,
                           match="'foobar' is not of type 'integer'"):
            J(data)

    def test_no_definition_key_no_example(self):
        """
        This tests that if we don't pass a definition_key and the schema
        doesn't define an example at the root that it will attempt to
        resolve it from the properties if the root is an object.
        """
        schema_file = os.path.join(
            os.path.dirname(__file__), 'schema', 'annotation_no_example.yaml')
        J = json_schema_type(schema_file=schema_file)

        # Verify description and example were set as attributes based
        # on values from the loaded json schema.
        assert 'An annotation object.' == J.description
        assert {
            'annotation_id': 1,
            'auth': 'eb25f25becca416092752b0f457f1271',
            'more_id': 1,
            'name': 'Annotation',
            'url': 'https://upsight.com',
            'urls': ['https://upsight.com'],
        } == J.example
        assert isinstance(J.schema, ResourceSchema)

    def test_no_definition_key_missing_description(self):
        schema_file = os.path.join(
            os.path.dirname(__file__), 'schema', 'no_description.yaml')
        with pytest.raises(TypeSystemError,
                           match='Schema is missing a description.'):
            json_schema_type(schema_file=schema_file)

    def test_definition_key(self):
        schema_file = os.path.join(
            os.path.dirname(__file__), 'schema', 'annotation.yaml')
        J = json_schema_type(
            schema_file=schema_file, definition_key='annotation_id')
        # Verify description and example were set as attributes based
        # on values from the loaded json schema.
        assert 'Auto-increment ID.' == J.description
        assert 1 == J.example
        assert J.native_type is int
        assert 'integer' == J.json_type
        assert isinstance(J.schema, ResourceSchema)

        # no exception
        J(1)
        # Not an int
        expected = "'not an int' is not of type 'integer'"
        with pytest.raises(TypeSystemError, match=expected):
            J('not an int')

    def test_definition_key_resolve_array_of_object(self):
        """
        This tests that when the definition is an array of objects
        that we resolve the array and object properties properly.
        """
        schema_file = os.path.join(
            os.path.dirname(__file__), 'schema', 'annotation_no_example.yaml')
        J = json_schema_type(
            schema_file=schema_file, definition_key='annotations')
        # Verify description and example were set as attributes based
        # on values from the loaded json schema.
        assert 'A list of annotation objects.' == J.description
        assert 'array' == J.json_type
        assert list == J.native_type
        assert [{
            'annotation_id': 1,
            'auth': 'eb25f25becca416092752b0f457f1271',
            'more_id': 1,
            'name': 'Annotation',
            'url': 'https://upsight.com',
            'urls': ['https://upsight.com']
        }] == J.example

    def test_definition_key_missing(self):
        schema_file = os.path.join(
            os.path.dirname(__file__), 'schema', 'annotation.yaml')
        expected = "Definition `does_not_exist` is not defined in the schema."
        with pytest.raises(TypeSystemError, match=expected):
            json_schema_type(
                schema_file=schema_file, definition_key='does_not_exist')

    def test_definition_key_missing_description(self):
        schema_file = os.path.join(
            os.path.dirname(__file__), 'schema', 'no_description.yaml')
        expected = 'Definition `annotation_id` is missing a description.'
        with pytest.raises(TypeSystemError, match=expected):
            json_schema_type(
                schema_file=schema_file, definition_key='annotation_id')

    def test_definition_key_ref_missing_description(self):
        """
        Tests if the definition is a reference that once resolved does not
        have a description, that we raise a TypeSystemError.
        """
        schema_file = os.path.join(
            os.path.dirname(__file__), 'schema', 'annotation.yaml')
        expected = "Definition `dog` is missing a description."
        with pytest.raises(TypeSystemError, match=expected):
            json_schema_type(
                schema_file=schema_file, definition_key='dog')

    def test_definition_key_bad_ref(self):
        """
        Tests when the desire definition is a reference that can't be resolved.
        """
        schema_file = os.path.join(
            os.path.dirname(__file__), 'schema', 'annotation.yaml')
        expected = "Unresolvable JSON pointer: 'definitions/doesnotexist'"
        with pytest.raises(TypeSystemError, match=expected):
            json_schema_type(
                schema_file=schema_file, definition_key='bad_ref')

    def test_definition_array_of_types(self):
        schema_file = os.path.join(
            os.path.dirname(__file__), 'schema', 'annotation.yaml')
        J = json_schema_type(schema_file, definition_key='name')
        assert 'string' == J.json_type
        assert str == J.native_type


def test_new_type_uses_parent_description():
    S = string('A string', example='Foo')
    N = new_type(S, nullable=True)
    assert 'A string' == N.description
    assert S.nullable is False
    assert N.nullable is True



from typing import Union


class DoctorError(ValueError):
    """Base error class for Doctor."""
    def __init__(self, message, errors: dict = None):
        self.errors = errors
        super().__init__(message)


#: Alias for DoctorError, for backwards compatibility.
SchematicError = DoctorError


class ForbiddenError(DoctorError):
    """Raised when a request is forbidden for the authorized user.

    Corresponds to a HTTP 403 Forbidden error.
    """
    pass


class ImmutableError(DoctorError):
    """Raised for immutable errors for a schema.

    Corresponds to a HTTP 409 Conflict error.
    """
    pass


class InternalError(DoctorError):
    """Raised when there is an internal server error.

    Corresponds to a HTTP 500 Internal Server Error.
    """
    pass


class InvalidValueError(DoctorError):
    """
    Raised for errors when doing more complex validation that
    can't be done in a schema.

    Corresponds to a HTTP 400 Bad Request error.
    """
    pass


class NotFoundError(DoctorError):
    """Raised when a resource is not found.

    Corresponds to a HTTP 404 Not Found error.
    """
    pass


class ParseError(DoctorError):
    """Raised when a value cannot be parsed into an appropriate type."""
    pass


class SchemaError(DoctorError):
    """Raised for errors in a schema."""
    pass


class SchemaLoadingError(DoctorError):
    """Raised when loading a resource and it is invalid."""
    pass


class SchemaValidationError(DoctorError):
    """Raised for errors when validating things against a schema."""

    def __init__(self, message, errors=None):
        super(SchemaValidationError, self).__init__(message)
        self.errors = errors


class TypeSystemError(DoctorError):
    """An error that represents an invalid value for a type.

    This is borrowed from apistar:
    https://github.com/encode/apistar/blob/master/apistar/exceptions.py#L1-L15

    :param detail: Detail about the error.
    :param cls: The class type that was being instantiated.
    :param code: The error code.
    :param errors: A dict containing all validation errors during the request.
        The key is the param name and the value is the error message.
    """
    def __init__(self,
                 detail: Union[str, dict] = None,
                 cls: type = None,
                 code: str = None,
                 errors: dict = None) -> None:

        if cls is not None and code is not None:
            cls_errors = getattr(cls, 'errors')
            detail = cls_errors[code].format(**cls.__dict__)

        self.detail = detail
        if errors and len(errors) == 1:
            param = list(errors.keys())[0]
            msg = list(errors.values())[0]
            detail = '{} - {}'.format(param, msg)
        super().__init__(detail, errors=errors)


class UnauthorizedError(DoctorError):
    """Raised when a request is unauthorized.

    Corresponds to a HTTP 401 Unauthorized error.
    """
    pass



import os

import jsonschema
import mock
import simplejson as json

from doctor.errors import (
    ParseError, SchemaError, SchemaLoadingError, SchemaValidationError)
from doctor.schema import Schema, SchemaRefResolver
from .base import TestCase


class TestSchema(TestCase):

    def setUp(self):
        self.schema = Schema.from_file(os.path.join(
            os.path.dirname(__file__), 'schema', 'annotation.yaml'))

    @mock.patch('doctor.schema.logging')
    def test_from_file_error(self, mock_logging):
        schema_filename = os.path.join(os.path.dirname(__file__),
                                       'schema', 'bad_yaml.yaml')
        with self.assertRaises(SchemaLoadingError):
            Schema.from_file(schema_filename)
        self.assertEqual(mock_logging.exception.call_args_list, [mock.call(
            'Error loading schema file {}'.format(schema_filename))])

    def test_get_validator(self):
        validator = self.schema.get_validator()
        self.assertIsInstance(validator, jsonschema.Draft4Validator)
        self.assertEqual(validator.resolver, self.schema.resolver)
        self.assertEqual(validator.schema, self.schema.schema)

    def test_resolver(self):
        self.assertIsNone(self.schema._resolver)
        resolver = self.schema.resolver
        self.assertIsInstance(resolver, jsonschema.RefResolver)
        self.assertEqual(resolver, self.schema._resolver)

    def test_validate(self):
        value = {'annotation_id': 1,
                 'name': 'hodor'}
        result = self.schema.validate(value, self.schema.get_validator())
        self.assertEqual(result, value)

    def test_validate_error(self):
        bad_value = {'annotation_id': 'hodor',
                     'name': 'hodor'}
        with self.assertRaisesRegexp(
                SchemaValidationError, r"'hodor' is not of type 'integer'"):
            self.schema.validate(bad_value, self.schema.get_validator())

    def test_validate_format_error(self):
        bad_value = {
            'annotation_id': 1,
            'name': 'hodor',
            'url': 'not-a-url-but-i-am-a-string',
        }
        with self.assertRaisesRegexp(
                SchemaValidationError,
                r"'not-a-url-but-i-am-a-string' is not a 'uri'"):
            self.schema.validate(bad_value, self.schema.get_validator())

    def test_validate_json(self):
        value = {'annotation_id': 1,
                 'name': 'hodor'}
        result = self.schema.validate_json(json.dumps(value),
                                           self.schema.get_validator())
        self.assertEqual(result, value)

    def test_validate_json_with_key_from_ref(self):
        value = {'annotation_id': 1,
                 'name': 'hodor',
                 'auth': 'abcd'}
        result = self.schema.validate_json(json.dumps(value),
                                           self.schema.get_validator())
        self.assertEqual(result, value)

    def test_validate_json_with_key_from_ref_invalid(self):
        value = {'annotation_id': 1,
                 'name': 'hodor',
                 'auth': 1}
        with self.assertRaisesRegexp(
                SchemaValidationError, r"1 is not of type 'string'"):
            self.schema.validate_json(json.dumps(value),
                                      self.schema.get_validator())

    def test_validate_json_error_invalid_data_multiple_errors(self):
        bad_value = json.dumps({'annotation_id': 'hodor',
                                'name': 1})
        expected_error = r"'hodor' is not of type 'integer'"
        with self.assertRaisesRegexp(
                SchemaValidationError, expected_error) as context:
            self.schema.validate_json(bad_value, self.schema.get_validator())
        errors = context.exception.errors
        expected = {
            'annotation_id': expected_error,
            'name': "1 is not of type 'null', 'string'",
        }
        self.assertEqual(expected, errors)

    def test_validate_json_error_invalid_json(self):
            with self.assertRaisesRegexp(ParseError, r'Error parsing JSON'):
                self.schema.validate_json(
                    'bad json', self.schema.get_validator())


class TestSchemaRefResolver(TestCase):

    def setUp(self):
        self.schema_filepath = os.path.join(os.path.dirname(__file__),
                                            'schema', 'annotation.yaml')
        self.base_uri = 'file://{}/'.format(
            os.path.dirname(os.path.abspath(self.schema_filepath)))
        self.schema = Schema.from_file(self.schema_filepath)
        self.resolver = self.schema.resolver
        self.assertIsInstance(self.resolver, SchemaRefResolver)

    def test_resolve(self):
        uri, value = self.resolver.resolve('#/test_ref')
        self.assertEqual(uri, self.base_uri + '#/definitions/annotation_id')
        self.assertEqual(value, {'description': 'Auto-increment ID.',
                                 'type': 'integer', 'example': 1})

    def test_resolve_external(self):
        """Should be able to resolve references to other files."""
        uri, value = self.resolver.resolve('#/properties/auth')
        self.assertEqual(uri, self.base_uri + 'common.yaml#/definitions/auth')
        self.assertEqual(value, {'description': 'auth token',
                                 'example': 'eb25f25becca416092752b0f457f1271',
                                 'type': ['string']})

    def test_resolve_error(self):
        with self.assertRaisesRegexp(
                SchemaError, r"Unresolvable JSON pointer: u?'invalid_ref'"):
            self.resolver.resolve('#/invalid_ref')

    def test_resolve_error_chain(self):
        expected_message = (
            r"Unresolvable JSON pointer: u?'invalid_ref_chain_3' \(from "
            r"/ => /#/invalid_ref_chain_1 => /#/invalid_ref_chain_2\)")
        with self.assertRaisesRegexp(SchemaError, expected_message):
            self.resolver.resolve('#/invalid_ref_chain_1')

    def test_resolve_error_circular_chain(self):
        """If there is a circular reference, should raise an error."""
        expected_message = (
            r"Circular reference in schema: / => /#/circular_ref_chain_1 => "
            r"/#/circular_ref_chain_2 => /#/circular_ref_chain_3 => "
            r"/#/circular_ref_chain_1")
        with self.assertRaisesRegexp(SchemaError, expected_message):
            self.resolver.resolve('#/circular_ref_chain_1')



import inspect

from flask_restful import Resource

from doctor.flask import handle_http
from doctor.routing import (
    create_routes, delete, get, get_handler_name, post, put, HTTPMethod, Route)
from doctor.utils import Params

from .types import Age, Foo, FooId, FooInstance, Foos, IsAlive, Name


def delete_foo(foo_id: FooId):
    pass


def get_foo(name: Name, age: Age, is_alive: IsAlive=True) -> Foo:
    return ''


def get_foos(is_alive: IsAlive=True) -> Foos:
    return ''


def create_foo(name: Name) -> Foo:
    return ''


def update_foo(foo_id: FooId, name: Name, is_alive: IsAlive=True) -> Foo:
    return ''


class TestRouting(object):

    def test_httpmethod(self):
        m = HTTPMethod('get', get_foo, allowed_exceptions=[ValueError],
                       title='Retrieve')
        assert 'get' == m.method
        assert inspect.signature(get_foo) == m.logic._doctor_signature
        expected = Params(
            all=['name', 'age', 'is_alive'],
            optional=['is_alive'],
            required=['name', 'age'],
            logic=['name', 'age', 'is_alive'])
        assert expected == m.logic._doctor_params
        assert [ValueError] == m.logic._doctor_allowed_exceptions
        assert 'Retrieve' == m.logic._doctor_title
        assert m.logic._doctor_req_obj_type is None

    def test_httpmethod_with_req_obj_type(self):
        m = HTTPMethod('get', get_foo, allowed_exceptions=[ValueError],
                       title='Retrieve', req_obj_type=FooInstance)
        assert 'get' == m.method
        assert inspect.signature(get_foo) == m.logic._doctor_signature
        expected = Params(
            all=['foo', 'foo_id'],
            optional=['foo'],
            required=['foo_id'],
            logic=['foo', 'foo_id'])
        assert expected == m.logic._doctor_params
        assert [ValueError] == m.logic._doctor_allowed_exceptions
        assert 'Retrieve' == m.logic._doctor_title
        assert FooInstance == m.logic._doctor_req_obj_type

    def test_delete(self):
        expected = HTTPMethod('delete', get_foo,
                              allowed_exceptions=[ValueError], title='Title')
        a = delete(get_foo, allowed_exceptions=[ValueError], title='Title')
        assert a.method == expected.method
        assert a.logic._doctor_title == 'Title'
        assert a.logic._doctor_allowed_exceptions == [ValueError]

    def test_get(self):
        expected = HTTPMethod('get', get_foo, allowed_exceptions=[ValueError],
                              title='Title')
        actual = get(get_foo, allowed_exceptions=[ValueError], title='Title')
        assert actual.method == expected.method
        assert actual.logic._doctor_title == 'Title'
        assert actual.logic._doctor_allowed_exceptions == [ValueError]

    def test_post(self):
        expected = HTTPMethod('post', get_foo, allowed_exceptions=[ValueError],
                              title='Title')
        actual = post(get_foo, allowed_exceptions=[ValueError], title='Title')
        assert actual.method == expected.method
        assert actual.logic._doctor_title == 'Title'
        assert actual.logic._doctor_allowed_exceptions == [ValueError]

    def test_put(self):
        expected = HTTPMethod('put', get_foo, allowed_exceptions=[ValueError],
                              title='Title')
        actual = put(get_foo, allowed_exceptions=[ValueError], title='Title')
        assert actual.method == expected.method
        assert actual.logic._doctor_title == 'Title'
        assert actual.logic._doctor_allowed_exceptions == [ValueError]

    def test_create_routes(self):
        class MyHandler(Resource):
            pass

        routes = (
            Route('^/foo/?$', (
                get(get_foos, title='Retrieve List'),
                post(create_foo)), base_handler_class=MyHandler,
                handler_name='MyHandler', heading='Foo'),
            Route('^/foo/<int:foo_id>/?$', (
                delete(delete_foo),
                get(get_foo),
                put(update_foo)), heading='Foo'),
            Route('^/foos/?$', (
                put(lambda: 'put'),), heading='Foo'),
        )
        actual = create_routes(routes, handle_http, Resource)

        # 2 routes created
        assert 3 == len(actual)

        # verify the first route
        route, handler = actual[0]
        assert r'^/foo/?$' == route

        # verify it's an instance of our base handler class.
        assert issubclass(handler, MyHandler)
        # verify it used our custom handler name
        assert 'MyHandler' == handler.__name__

        # verify each http method was added
        assert hasattr(handler, 'get')
        assert hasattr(handler, 'post')

        # verify heading attr was added to handler
        assert handler._doctor_heading == 'Foo'

        # verify params for get
        params = handler.get._doctor_params
        expected = Params(
            all=['is_alive'], required=[], optional=['is_alive'],
            logic=['is_alive'])
        assert expected == params

        # verify signature
        sig = handler.get._doctor_signature
        expected = inspect.signature(get_foos)
        assert expected == sig

        # verify custom title
        assert 'Retrieve List' == handler.get._doctor_title

        # verify params for post
        params = handler.post._doctor_params
        expected = Params(
            all=['name'], required=['name'], optional=[], logic=['name'])
        assert expected == params

        # verify signature
        sig = handler.post._doctor_signature
        expected = inspect.signature(create_foo)
        assert expected == sig

        # verify the second route
        route, handler = actual[1]
        assert '^/foo/<int:foo_id>/?$' == route
        # verify each http method was added
        assert hasattr(handler, 'get')
        assert hasattr(handler, 'delete')
        assert hasattr(handler, 'put')
        # verify it generated an appropriate class handler name
        assert 'FooHandler' == handler.__name__

        # Verify the 3rd handler which would have had a conflicting handler
        # name has a number appended to the end of it.
        route, handler = actual[2]
        assert 'FooHandler2' == handler.__name__

    def test_create_routes_reuse_logic_different_title(self):
        """
        This tests that if we want to reuse a logic function with a different
        route, and give it a unique title that it persists.  Previously if you
        defined a title for 2 different routes using the same logic function,
        it would use the last defined title on both routes.
        """
        routes = (
            Route('^/foo/?$', methods=[
                get(get_foos, title='Retrieve List')],
                heading='Foo'
            ),
            Route('^/bar/?$', methods=[
                get(get_foos, title='Retrieve Other List')],
                heading='Foo'
            ),
        )
        actual = create_routes(routes, handle_http, Resource)

        # verify the first route
        route, handler = actual[0]
        assert r'^/foo/?$' == route
        assert 'Retrieve List' == handler.get._doctor_title

        # verify the second route
        route, handler = actual[1]
        assert r'^/bar/?$' == route
        assert 'Retrieve Other List' == handler.get._doctor_title

    def test_get_handler_name_route_has_handler_name(self):
        """Tests handler name comes from one defined on Route"""
        route = Route('/', (get(get_foos),), handler_name='FooFooHandler')
        assert 'FooFooHandler' == get_handler_name(route, get_foos)

    def test_get_handler_name_from_logic_function_name(self):
        """Tests handler names is generated from logic func name."""
        route = Route('/', (get(get_foos),))
        assert 'GetFoosHandler' == get_handler_name(route, get_foos)

    def test_get_handler_name_heading_list_handler(self):
        """
        Tests that we get the handler name from the defined heading and
        http methods.  Since post is included it infers it's a list endpoint.
        """
        route = Route('/', (post(create_foo),), heading='Dinosaur (v1)')
        assert 'DinosaurV1ListHandler' == get_handler_name(route, create_foo)

    def test_get_handler_name_list_handler(self):
        """
        Tests that we get the handler name using the logic function and also
        including List in the name since it contains a post method.
        """
        route = Route('/', (post(create_foo),))
        assert 'CreateFooListHandler' == get_handler_name(route, create_foo)

    def test_get_handler_name_route_heading(self):
        """
        Tests that we get the handler name using the route heading for a non
        list endpoint.
        """
        route = Route('/', (put(update_foo),), heading='Dinosaur (v1)')
        assert 'DinosaurV1Handler' == get_handler_name(route, update_foo)



import pytest

from doctor.docs.base import BaseHarness
from doctor.resource import ResourceAnnotation

from .types import Age, Color, ItemId, Name
from .utils import add_doctor_attrs


@pytest.fixture
def annotation():
    def logic(age: Age, name: Name, item_id: ItemId=None, color: Color='green'):
        pass

    logic = add_doctor_attrs(logic)
    annotation = ResourceAnnotation(logic, 'GET')
    return annotation


def test_get_example_values(annotation):
    harness = BaseHarness('/api')
    route = '/foo'
    actual = harness._get_example_values(route, annotation)
    expected = {
        'age': 34,
        'color': 'blue',
        'item_id': 1,
        'name': 'John',
    }
    assert expected == actual



import json
import os

import mock

from doctor.docs.flask import AutoFlaskHarness

from .base import TestCase


class TestDocsFlask(TestCase):

    def setUp(self):
        flask_folder = os.path.join(os.path.dirname(__file__))
        self.harness = AutoFlaskHarness(
            os.path.join(flask_folder, 'flask_app.py'), 'http://127.0.0.1/')
        self.harness.setup_app(mock.sentinel.sphinx_app)
        self.annotations = list(self.harness.iter_annotations())

    def tearDown(self):
        self.harness.teardown_app(mock.sentinel.sphinx_app)

    def test_harness_iter_annotations(self):
        self.assertEqual(len(self.annotations), 4)

        heading, rule, view_class, annotations = self.annotations[0]
        self.assertEqual(heading, 'API Status')
        self.assertEqual(rule.rule, '/')
        self.assertEqual([annotation.http_method for annotation in annotations],
                         ['GET'])

        heading, rule, view_class, annotations = self.annotations[1]
        self.assertEqual(heading, 'Notes (v1)')
        self.assertEqual(rule.rule, '/note/')
        self.assertEqual([annotation.http_method for annotation in annotations],
                         ['GET', 'POST'])

        heading, rule, view_class, annotations = self.annotations[2]
        self.assertEqual(heading, 'Notes (v1)')
        self.assertEqual(rule.rule, '/note/<int:note_id>/')
        self.assertEqual([annotation.http_method for annotation in annotations],
                         ['GET', 'PUT', 'DELETE'])

        heading, rule, view_class, annotations = self.annotations[3]
        assert 'Z' == heading
        assert '/example/list-obj/' == rule.rule
        assert ['GET'] == [annotation.http_method for annotation in annotations]

    def test_harness_request_get_query_string(self):
        _, rule, view_class, annotations = self.annotations[2]
        annotation = annotations[0]
        result = self.harness.request(rule, view_class, annotation)
        result['response'] = json.loads(result['response'])
        self.assertEqual(result, {
            'method': 'GET',
            'params': {},
            'response': {'body': 'Example body',
                         'done': True,
                         'note_id': 1},
            'url': 'http://127.0.0.1/note/1/?note_type=quick'})

    def test_harness_request_get_list_and_dict_params(self):
        """
        This test verifies that we json dumps object and array types when
        building the url's query string parameters if this is a GET request.
        """
        _, rule, view_class, annotations = self.annotations[3]
        annotation = annotations[0]
        result = self.harness.request(rule, view_class, annotation)
        result['response'] = json.loads(result['response'])
        expected = {

            'method': 'GET',
            'params': {},
            'response': {},
            'url': ('http://127.0.0.1/example/list-obj/?note_types=%5B%22quick'
                    '%22%5D&a_note=%7B%22body%22%3A+%22Example+Body%22%2C+%22'
                    'done%22%3A+true%2C+%22note_id%22%3A+1%7D')
        }
        self.assertEqual(expected, result)

    def test_harness_request_post(self):
        _, rule, view_class, annotations = self.annotations[1]
        annotation = annotations[1]
        result = self.harness.request(rule, view_class, annotation)
        result['response'] = json.loads(result['response'])
        self.assertEqual(result, {
            'method': 'POST',
            'params': {'body': 'body',
                       'done': False},
            'response': {'body': 'body',
                         'done': False,
                         'note_id': 2},
            'url': 'http://127.0.0.1/note/'})

    def test_harness_request_delete(self):
        response_mock = mock.Mock(data='')
        self.harness.test_client.delete = mock.Mock(return_value=response_mock)
        _, rule, view_class, annotations = self.annotations[2]
        annotation = annotations[2]
        result = self.harness.request(rule, view_class, annotation)

        expected_url = '/note/1/'
        self.harness.test_client.delete.assert_called_with(
            expected_url, data={}, headers={})
        expected_result = {
            'url': 'http://127.0.0.1/note/1/',
            'params': {},
            'method': 'DELETE',
            'response': ''
        }
        self.assertEqual(expected_result, result)

    def test_harness_define_example_values(self):
        _, rule, view_class, annotations = self.annotations[1]
        annotation = annotations[1]
        result = self.harness.request(rule, view_class, annotation)
        result['response'] = json.loads(result['response'])
        self.assertEqual(result, {
            'method': 'POST',
            'params': {'body': 'body',
                       'done': False},
            'response': {'body': 'body',
                         'done': False,
                         'note_id': 2},
            'url': 'http://127.0.0.1/note/'})

        # Now override the values and validate that they're different
        self.harness.define_example_values('POST', '/note/', {
            'body': 'This is a replaced body.'
        })
        result = self.harness.request(rule, view_class, annotation)
        result['response'] = json.loads(result['response'])
        self.assertEqual(result['params'], {'body': 'This is a replaced body.'})

        # Override, but with update=True, which should merge the values
        self.harness.define_example_values('POST', '/note/', {
            'body': 'This is an updated body.'
        }, update=True)
        result = self.harness.request(rule, view_class, annotation)
        result['response'] = json.loads(result['response'])
        self.assertEqual(result['params'], {'body': 'This is an updated body.',
                                            'done': False})



"""
This module provides Sphinx directives to generate documentation for Flask
resources which have been annotated with schema information. It is broken into
a separate module so that Flask applications using doctor for
validation don't need to include Sphinx in their runtime dependencies.
"""

from __future__ import absolute_import

import json
from collections import defaultdict
from urllib import parse

from ..resource import ResourceAnnotation
from ..utils import get_module_attr
from .base import BaseDirective, BaseHarness, HTTP_METHODS


class AutoFlaskDirective(BaseDirective):

    """Sphinx directive to document schema annotated Flask resources."""

    directive_name = 'autoflask'


class AutoFlaskHarness(BaseHarness):

    def __init__(self, app_module_filename, url_prefix):
        super(AutoFlaskHarness, self).__init__(url_prefix)
        self.app_module_filename = app_module_filename

    def __getstate__(self):  # pragma: no cover
        state = self.__dict__.copy()
        del state['app']
        del state['test_client']
        return state

    def __setstate__(self, state):  # pragma: no cover
        self.__dict__.update(state)

    def iter_annotations(self):
        """Yield a tuple for each Flask handler containing annotated methods.

        Each tuple contains a heading, routing rule, the view class associated
        with the rule, and the annotations for the methods in that class.
        """
        # Need to store a list of route, view_class, and annotations by a
        # section key so that all methods of a resource are kept together in
        # the documentation.  The key of the section will be the heading that
        # the route documentation goes under.
        section_map = defaultdict(list)
        for rule in self.app.url_map.iter_rules():
            if rule.endpoint == 'static':
                # Don't document static file endpoints.
                continue
            # This gives us the auto-generated view function.
            view_function = self.app.view_functions.get(rule.endpoint)
            if view_function is None:
                continue
            # This gives us the actual Flask resource class.
            view_class = getattr(view_function, 'view_class', None)
            if view_class is None:
                continue

            annotations = []
            for method_name in HTTP_METHODS:
                method = getattr(view_class, method_name, None)
                if not method:
                    continue
                annotation = ResourceAnnotation(
                    method, method_name, method._doctor_title)
                annotations.append(annotation)
            if annotations:
                heading = self._get_annotation_heading(view_class, str(rule))
                section_map[heading].append((rule, view_class, annotations))

        # Loop through each heading and it's items and yield the values.
        for heading in sorted(section_map.keys()):
            for item in section_map[heading]:
                rule, view_class, annotations = item
                yield (heading, rule, view_class, annotations)

    def request(self, rule, view_class, annotation):
        """Make a request against the app.

        This attempts to use the schema to replace any url params in the path
        pattern. If there are any unused parameters in the schema, after
        substituting the ones in the path, they will be sent as query string
        parameters or form parameters. The substituted values are taken from
        the "example" value in the schema.

        Returns a dict with the following keys:

        - **url** -- Example URL, with url_prefix added to the path pattern,
          and the example values substituted in for URL params.
        - **method** -- HTTP request method (e.g. "GET").
        - **params** -- A dictionary of query string or form parameters.
        - **response** -- The text response to the request.

        :param route: Werkzeug Route object.
        :param view_class: View class for the annotated method.
        :param annotation: Annotation for the method to be requested.
        :type annotation: doctor.resource.ResourceAnnotation
        :returns: dict
        """
        headers = self._get_headers(rule, annotation)
        example_values = self._get_example_values(rule, annotation)
        # If any of the example values for DELETE/GET HTTP methods are dicts
        # or lists, we will need to json dump them before building the rule,
        # otherwise the query string parameter won't get parsed correctly
        # by doctor.
        if annotation.http_method.upper() in ('DELETE', 'GET'):
            for key, value in list(example_values.items()):
                if isinstance(value, (dict, list)):
                    example_values[key] = json.dumps(value)
        _, path = rule.build(example_values, append_unknown=True)
        if annotation.http_method.upper() not in ('DELETE', 'GET'):
            parsed_path = parse.urlparse(path)
            path = parsed_path.path
            params = example_values
        else:
            params = {}
        method_name = annotation.http_method.lower()
        method = getattr(self.test_client, method_name)
        if method_name in ('post', 'put'):
            response = method(path, data=json.dumps(params), headers=headers,
                              content_type='application/json')
        else:
            response = method(path, data=params, headers=headers)
        return {
            'url': '/'.join([self.url_prefix, path.lstrip('/')]),
            'method': annotation.http_method.upper(),
            'params': params,
            'response': response.data,
        }

    def setup_app(self, sphinx_app):
        self.app = get_module_attr(self.app_module_filename, 'app', {})
        self.test_client = self.app.test_client()


def setup(app):  # pragma: no cover
    """This setup function is called by Sphinx."""
    AutoFlaskDirective.setup(app)



import functools
import inspect
import logging
import os
import re
import sys
import types
from copy import copy
from inspect import Parameter, Signature
from typing import Callable, List

try:
    from sphinx.util.docstrings import prepare_docstring
except ImportError:
    prepare_docstring = None

from doctor.types import SuperType

#: Used to identify the end of the description block, and the beginning of the
#: parameters. This assumes that the parameters and such will always occur at
#: the end of the docstring.
DESCRIPTION_END_RE = re.compile(':(arg|param|returns|throws)', re.I)


def copy_func(func: Callable) -> Callable:
    """Returns a copy of a function.

    :param func: The function to copy.
    :returns: The copied function.
    """
    copied = types.FunctionType(
        func.__code__, func.__globals__, name=func.__name__,
        argdefs=func.__defaults__, closure=func.__closure__)
    copied = functools.update_wrapper(copied, func)
    copied.__kwdefaults__ = func.__kwdefaults__
    return copied


class RequestParamAnnotation(object):
    """Represents a new request parameter annotation.

    :param name: The name of the parameter.
    :param annotation: The annotation type of the parameter.
    :type annotation: A doctor type that should subclass
        :class:`~doctor.types.SuperType`.
    :param required: Indicates if the parameter is required or not.
    """
    def __init__(self, name: str, annotation, required: bool=False):
        self.annotation = annotation
        self.name = name
        self.required = required


class Params(object):
    """Represents parameters for a request.

    :param all: A list of all paramter names for a request.
    :param required: A list of all required parameter names for a request.
    :param optional: A list of all optional parameter names for a request.
    :param logic: A list of all parameter names that are part ofthe logic
        function signature.
    """
    def __init__(self, all: List[str], required: List[str],
                 optional: List[str], logic: List[str]):
        self.all = all
        self.optional = optional
        self.required = required
        self.logic = logic

    def __repr__(self):
        return str({
            'all': self.all,
            'logic': self.logic,
            'optional': self.optional,
            'required': self.required
        })

    def __eq__(self, other):
        for attr in ('all', 'logic', 'optional', 'required'):
            if getattr(self, attr) != getattr(other, attr):
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)


def get_params_from_func(func: Callable, signature: Signature=None) -> Params:
    """Gets all parameters from a function signature.

    :param func: The function to inspect.
    :param signature: An inspect.Signature instance.
    :returns: A named tuple containing information about all, optional,
        required and logic function parameters.
    """
    if signature is None:
        # Check if the function already parsed the signature
        signature = getattr(func, '_doctor_signature', None)
        # Otherwise parse the signature
        if signature is None:
            signature = inspect.signature(func)

    # Check if a `req_obj_type` was provided for the function.  If so we should
    # derrive the parameters from that defined type instead of the signature.
    if getattr(func, '_doctor_req_obj_type', None):
        annotation = func._doctor_req_obj_type
        all_params = list(annotation.properties.keys())
        required = annotation.required
        optional = list(set(all_params) - set(required))
    else:
        # Required is a positional argument with no defualt value and it's
        # annotation must sub class SuperType.  This is so we don't try to
        # require parameters passed to a logic function by a decorator that are
        # not part of a request.
        required = [key for key, p in signature.parameters.items()
                    if p.default == p.empty and
                    issubclass(p.annotation, SuperType)]
        optional = [key for key, p in signature.parameters.items()
                    if p.default != p.empty]
        all_params = [key for key in signature.parameters.keys()]

    # Logic params are all parameters that are part of the logic signature.
    logic_params = copy(all_params)
    return Params(all_params, required, optional, logic_params)


def add_param_annotations(
        logic: Callable, params: List[RequestParamAnnotation]) -> Callable:
    """Adds parameter annotations to a logic function.

    This adds additional required and/or optional parameters to the logic
    function that are not part of it's signature.  It's intended to be used
    by decorators decorating logic functions or middleware.

    :param logic: The logic function to add the parameter annotations to.
    :param params: The list of RequestParamAnnotations to add to the logic func.
    :returns: The logic func with updated parameter annotations.
    """
    # If we've already added param annotations to this function get the
    # values from the logic, otherwise we need to inspect it.
    if hasattr(logic, '_doctor_signature'):
        sig = logic._doctor_signature
        doctor_params = logic._doctor_params
    else:
        sig = inspect.signature(logic)
        doctor_params = get_params_from_func(logic, sig)

    prev_parameters = {name: param for name, param in sig.parameters.items()}
    new_params = []
    for param in params:
        # If the parameter already exists in the function signature, log
        # a warning and skip it.
        if param.name in prev_parameters:
            logging.warning('Not adding %s to signature of %s, function '
                            'already has that parameter in its signature.',
                            param.name, logic.__name__)
            continue
        doctor_params.all.append(param.name)
        default = None
        if param.required:
            default = Parameter.empty
            doctor_params.required.append(param.name)
        else:
            doctor_params.optional.append(param.name)
        new_params.append(
            Parameter(param.name, Parameter.KEYWORD_ONLY, default=default,
                      annotation=param.annotation))

    new_sig = sig.replace(
        parameters=list(prev_parameters.values()) + new_params)
    logic._doctor_signature = new_sig
    logic._doctor_params = doctor_params
    return logic


def get_module_attr(module_filename, module_attr, namespace=None):
    """Get an attribute from a module.

    This uses exec to load the module with a private namespace, and then
    plucks and returns the given attribute from that module's namespace.

    Note that, while this method doesn't have any explicit unit tests, it is
    tested implicitly by the doctor's own documentation. The Sphinx
    build process will fail to generate docs if this does not work.

    :param str module_filename: Path to the module to execute (e.g.
        "../src/app.py").
    :param str module_attr: Attribute to pluck from the module's namespace.
        (e.g. "app").
    :param dict namespace: Optional namespace. If one is not passed, an empty
        dict will be used instead. Note that this function mutates the passed
        namespace, so you can inspect a passed dict after calling this method
        to see how the module changed it.
    :returns: The attribute from the module.
    :raises KeyError: if the module doesn't have the given attribute.
    """
    if namespace is None:
        namespace = {}
    module_filename = os.path.abspath(module_filename)
    namespace['__file__'] = module_filename
    module_dir = os.path.dirname(module_filename)
    old_cwd = os.getcwd()
    old_sys_path = sys.path[:]
    try:
        os.chdir(module_dir)
        sys.path.append(module_dir)
        with open(module_filename, 'r') as mf:
            exec(compile(mf.read(), module_filename, 'exec'), namespace)
        return namespace[module_attr]
    finally:
        os.chdir(old_cwd)
        sys.path = old_sys_path


def get_description_lines(docstring):
    """Extract the description from the given docstring.

    This grabs everything up to the first occurrence of something that looks
    like a parameter description. The docstring will be dedented and cleaned
    up using the standard Sphinx methods.

    :param str docstring: The source docstring.
    :returns: list
    """
    if prepare_docstring is None:
        raise ImportError('sphinx must be installed to use this function.')

    if not isinstance(docstring, str):
        return []
    lines = []
    for line in prepare_docstring(docstring):
        if DESCRIPTION_END_RE.match(line):
            break
        lines.append(line)
    if lines and lines[-1] != '':
        lines.append('')
    return lines


def get_valid_class_name(s: str) -> str:
    """Return the given string converted so that it can be used for a class name

    Remove leading and trailing spaces; removes spaces and capitalizes each
    word; and remove anything that is not alphanumeric.  Returns a pep8
    compatible class name.

    :param s: The string to convert.
    :returns: The updated string.
    """
    s = str(s).strip()
    s = ''.join([w.title() for w in re.split('\W+|_', s)])
    return re.sub(r'[^\w|_]', '', s)



import os
from setuptools import setup

# Use exec so pip can get the version before installing the module
version_filename = os.path.abspath(os.path.join(
    os.path.dirname(__file__), 'doctor', '_version.py'))
with open(version_filename, 'r') as vf:
    exec(compile(vf.read(), version_filename, 'exec'), globals(), locals())

readme_path = os.path.abspath(
    os.path.join(os.path.dirname(__file__), 'README.rst'))
with open(readme_path, 'r') as fp:
    long_description = fp.read()

setup(
    name='doctor',
    version=__version__,  # noqa -- flake8 should ignore this line
    description=('This module uses python 3 type hints to validate request and '
                 'response data in Flask Python APIs and generate API '
                 'documentation.'),
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://github.com/upsight/doctor',
    license='MIT',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
    ],
    keywords=['python', 'flask', 'json', 'api', 'validation',
              'documentation', 'sphinx'],
    author='Upsight',
    author_email='dev@upsight.com',
    packages=[
        'doctor',
        'doctor.docs',
        'doctor.utils',
    ],
    install_requires=[
        'isodate >= 0.6.0, < 1.0.0',
        'jsonschema >= 2.5.1, < 3.0.0',
        'pyyaml >= 3.11, < 4.0',
        'rfc3987 >= 1.3.4, < 2.0.0',
        'simplejson >= 3.6.3, < 4.0.0',
        'strict-rfc3339 >= 0.5, < 1.0',
    ],
    extras_require={
        'docs': [
            'mock >= 2.0.0, < 3.0.0',
            'sphinx >= 1.5.4, < 2.0.0',
            'sphinx-autodoc-typehints >= 1.2.4, < 2.0.0',
            'sphinx-rtd-theme >= 0.2.4, < 1.0.0',
            'sphinxcontrib-httpdomain >= 1.5.0, < 2.0.0',
        ],
        'tests': [
            'coverage >= 4.4.1, < 5.0.0',
            'flake8 >= 3.3.0, < 4.0.0',
            'flask >= 0.10.1, < 1.0.0',
            'flask-restful==0.3.6',
            'Flask-Testing==0.6.2',
            'mock >= 2.0.0, < 3.0.0',
            'pytest >= 3.3.2, < 4.0.0',
        ],
    },
)



import json

import mock
from werkzeug.routing import Rule

from doctor.docs import base
from doctor.resource import ResourceAnnotation
from doctor.response import Response

from .base import TestCase
from .types import (
    Age, Auth, Color, ExampleArray, ExampleObject, ExampleObjects, FooInstance,
    IsAlive, IsDeleted, Name)
from .utils import add_doctor_attrs


class TestDocsBase(TestCase):

    def test_prefix_lines_bytes(self):
        """
        This is a regression test where the response was a bytes instance.
        """
        lines = b'"Notes API v1.0.0"'
        prefix = '   '
        expected = ['   "Notes API v1.0.0"']
        assert expected == base.prefix_lines(lines, prefix)

    def test_get_example_lines_json(self):
        """Tests an example when the response is valid JSON."""
        headers = {'GeoIp-Country-Code': 'US'}
        lines = base.get_example_lines(headers, 'GET', 'http://example.com/',
                                       {}, json.dumps({'foo': 1, 'bar': 2}))
        self.assertEqual(lines, [
            '',
            'Example Request:',
            '',
            '.. code-block:: bash',
            '',
            '   curl http://example.com/ -X GET -H \'GeoIp-Country-Code: US\'',
            '',
            'Example Response:',
            '',
            '.. code-block:: json',
            '',
            '   {',
            '     "bar": 2,',
            '     "foo": 1',
            '   }',
        ])

    def test_get_example_lines_text(self):
        """Tests an example when the response is *not* valid JSON."""
        lines = base.get_example_lines({}, 'GET', 'http://example.com/', {},
                                       'hello, world!')
        self.assertEqual(lines, [
            '',
            'Example Request:',
            '',
            '.. code-block:: bash',
            '',
            '   curl http://example.com/ -X GET',
            '',
            'Example Response:',
            '',
            '.. code-block:: text',
            '',
            '   hello, world!',
        ])

    def test_get_json_object_lines_for_request_with_enum(self):
        def mock_logic(auth: Auth, is_alive: IsAlive, name: Name=None,
                       color: Color='blue'):
            pass

        mock_logic = add_doctor_attrs(mock_logic)
        annotation = ResourceAnnotation(mock_logic, 'GET')
        parameters = annotation.logic._doctor_signature.parameters
        properties = {k: p.annotation for k, p in parameters.items()}
        result = base.get_json_object_lines(
            annotation, properties, field='>json', url_params=[],
            request=True)
        assert result == [
            ':>json str auth: **Required**.  auth token',
            ':>json bool is_alive: **Required**.  Is alive?',
            ":>json str color: Color Must be one of: `['blue', 'green']`.",
            ':>json str name: name']

    def test_get_json_object_lines_for_request(self):
        """
        This tests that when the request kwarg is True that any
        required params have the description prefixed with
        **Required** and sorted in alphabetical order, followed by
        any optional parameters in alpabetical order.
        """
        def mock_logic(auth: Auth, age: Age, is_deleted: IsDeleted=True):
            pass

        mock_logic = add_doctor_attrs(mock_logic)
        annotation = ResourceAnnotation(mock_logic, 'GET')
        parameters = annotation.logic._doctor_signature.parameters
        properties = {k: p.annotation for k, p in parameters.items()}
        url_params = ['age']
        result = base.get_json_object_lines(
            annotation, properties, field='>json', url_params=url_params,
            request=True)
        self.assertEqual(result, [
            ':param int age: **Required**.  age',
            ':>json str auth: **Required**.  auth token',
            (':>json bool is_deleted: Indicates if the item should be marked '
             'as deleted'),
        ])

    def test_get_json_object_lines_object_response(self):
        """
        This tests that when our response is an object that we return
        all of it's documented properties.
        """
        def mock_logic() -> ExampleObject:
            pass

        mock_logic = add_doctor_attrs(mock_logic)
        annotation = ResourceAnnotation(mock_logic, 'GET')
        result = base.get_json_lines(
            annotation, field='>json', route='/foo', request=False)
        expected = [
            ':>json str str: auth token'
        ]
        self.assertEqual(expected, result)

    def test_get_json_lines_logic_defines_req_obj_type(self):
        """
        This tests that we properly generate the json params for a request
        when the logic function defines a `req_obj_type`.
        """
        def mock_logic(foo: FooInstance):
            pass

        mock_logic = add_doctor_attrs(mock_logic, req_obj_type=FooInstance)
        annotation = ResourceAnnotation(mock_logic, 'POST')
        result = base.get_json_lines(
            annotation, field='<json', route='/foo', request=True)
        expected = [
            ':<json int foo_id: **Required**.  foo id',
            ':<json str foo: foo'
        ]
        assert expected == result

    def test_get_json_lines_array_response(self):
        """
        Verifies we document properties of an array of objects.
        """
        def mock_logic() -> ExampleObjects:
            pass

        mock_logic = add_doctor_attrs(mock_logic)
        annotation = ResourceAnnotation(mock_logic, 'GET')
        result = base.get_json_lines(
            annotation, field='<json', route='/foo')
        self.assertEqual(result, [':<jsonarr str str: auth token'])

    def test_get_json_lines_response_response(self):
        """
        Verifies when our response is a doctor.response.Response instance
        and it has a type associated with it that we use that type to
        document it.
        """
        def mock_logic() -> Response[ExampleObject]:
            pass

        mock_logic = add_doctor_attrs(mock_logic)
        annotation = ResourceAnnotation(mock_logic, 'GET')
        result = base.get_json_lines(
            annotation, field='>json', route='/foo', request=False)
        expected = [
            ':>json str str: auth token'
        ]
        assert expected == result

    def test_get_name(self):
        mock_class = mock.Mock()
        mock_class.__module__ = 'foo.bar'
        mock_class.__name__ = 'baz'
        self.assertEqual(base.get_name(mock_class), 'foo.bar.baz')
        mock_class.__module__ = '__builtin__'
        self.assertEqual(base.get_name(mock_class), 'baz')


class TestDocsBaseHarness(TestCase):

    def test_init(self):
        harness = base.BaseHarness('http://foo/')
        self.assertEqual(harness.url_prefix, 'http://foo')

    def test_get_annotation_heading_doctor_heading(self):
        """
        This test verifies we use the _doctor_heading attribute of the
        handler if it is present.
        """
        handler = mock.Mock(_doctor_heading='Test Title')
        route = '^foo/?$'
        harness = base.BaseHarness('http://foo/')
        actual = harness._get_annotation_heading(handler, route)
        self.assertEqual('Test Title', actual)

    def test_get_annotation_heading_class_path(self):
        """
        This test verifies if the class path has a resource name in it,
        that we use it for the heading.
        e.g. class <api.handlers.foo_bar.FooListHandler> becomes `Foo Bar`
        """
        handler = mock.Mock(spec_set=base.BaseHarness)
        handler.__str__ = mock.Mock(
            return_value='<class api.foo_bar.FooBarListHandler>')
        route = '^foo_bar/?$'
        harness = base.BaseHarness('http://foo/')
        actual = harness._get_annotation_heading(handler, route)

        expected = 'Foo Bar'
        self.assertEqual(expected, actual)

    def test_get_annotation_heading_generic_handlers(self):
        """
        This test verifies if our handlers are not in their own resource
        modules that we get the heading from the handler class name.
        e.g. class <api.handlers.handlers.FooBarListHandler> becomes `Foo Bar`
        """
        handler = mock.Mock(spec_set=base.BaseHarness)
        handler.__str__ = mock.Mock(
            return_value='<class api.handlers.FooBarListHandler>')
        route = '^foo_bar/?$'

        harness = base.BaseHarness('http://foo/')
        actual = harness._get_annotation_heading(handler, route)

        expected = 'Foo Bar'
        self.assertEqual(expected, actual)

    def test_get_annotation_heading_class_name_only(self):
        """
        This test verifies that if our handler has no path and is just
        the class name that we get the heading from the name.
        e.g. class <FooBarListHandler> becomes `Foo Bar`
        """
        handler = mock.Mock(spec_set=base.BaseHarness)
        handler.__str__ = mock.Mock(
            return_value='<class FooBarListHandler>')
        route = '^foo_bar/?$'

        harness = base.BaseHarness('http://foo/')
        actual = harness._get_annotation_heading(handler, route)

        expected = 'Foo Bar'
        self.assertEqual(expected, actual)

    def test_get_annotation_heading_class_path_internal(self):
        """
        This test verifies the path where the class path has the resource
        name in it and it's an internal route.
        """
        handler = mock.Mock(spec_set=base.BaseHarness)
        handler.__str__ = mock.Mock(
            return_value='<class api.foo_bar.InternalFooBarListHandler>')
        route = '^internal/r/foo_bar/?$'
        harness = base.BaseHarness('http://foo/')
        actual = harness._get_annotation_heading(handler, route)

        expected = 'Foo Bar (Internal)'
        self.assertEqual(expected, actual)

    @mock.patch('doctor.docs.base.hasattr')
    def test_get_annotation_heading_class_name_only_internal(self, mock_has):
        """
        This test verifies the path where our handler has no path and just
        the class name and that class is internal.
        """
        mock_has.return_value = False
        handler = mock.MagicMock(spec_set=base.BaseHarness)
        handler.__str__ = mock.Mock(
            return_value='<class InternalFooBarListHandler>')
        route = '^/internal/foo_bar/?$'

        harness = base.BaseHarness('http://foo/')
        actual = harness._get_annotation_heading(handler, route)

        expected = 'Foo Bar (Internal)'
        self.assertEqual(expected, actual)

    def test_get_example_values_get_http_method_with_list_and_dict_vals(self):
        """
        This test verifies if the route we are generating example values for
        is a GET endpoint and the example values are lists or dicts, that we
        json.dumps them when they are returned.  If the HTTP method is any
        other method, they will be returned as normal lists or dicts.
        """
        def mock_logic(e: ExampleArray, f: ExampleObject):
            pass

        mock_logic = add_doctor_attrs(mock_logic)
        route = Rule('/foo/bar/')

        annotation = ResourceAnnotation(mock_logic, 'GET')
        harness = base.BaseHarness('http://foo/bar/')
        example_values = harness._get_example_values(route, annotation)
        expected = {
            'e': json.dumps(['ex', 'array']),
            'f': json.dumps({'str': 'ex str'}),
        }
        self.assertEqual(expected, example_values)

        # Change the http method to something other than GET, they should
        # not be json dumped.
        annotation.http_method = 'POST'
        example_values = harness._get_example_values(route, annotation)
        expected = {
            'e': ['ex', 'array'],
            'f': {'str': 'ex str'},
        }
        self.assertEqual(expected, example_values)

    def test_get_example_values_when_logic_defines_req_obj_type(self):
        """
        This tests that we generate example values appropriately when the
        route defineds a req_obj_type which will pass all request params as
        that object instance to the logic function.

        If a req_obj_type was not defined for the logic, it would expect
        the json body to look like:

        {
            "foo": {
                "foo": "foo",
                "foo_id": 1
            }
        }

        Defining a req_obj_type tells the code that the request body should
        contain those attributes rather than a sub-key within the request.
        """
        def mock_logic(foo: FooInstance):
            pass

        mock_logic = add_doctor_attrs(mock_logic, req_obj_type=FooInstance)
        route = Rule('/foo/bar/')

        annotation = ResourceAnnotation(mock_logic, 'POST')
        harness = base.BaseHarness('http://foo/bar/')
        example_values = harness._get_example_values(route, annotation)
        expected = {
            'foo': 'foo',
            'foo_id': 1,
        }
        assert expected == example_values

    def test_class_name_to_resource_name(self):
        tests = (
            # (input, expected)
            ('Foo', 'Foo'),
            ('FooBar', 'Foo Bar'),
            ('FooV1Bar', 'Foo V1 Bar'),
            ('Reallylongnamewithnoothercase', 'Reallylongnamewithnoothercase'),
            ('HTTPResponse', 'HTTP Response'),
        )
        for arg, expected in tests:
            assert expected == base.class_name_to_resource_name(arg)

    @mock.patch.dict('doctor.docs.base.ALL_RESOURCES',
                     {'An Object': ExampleObject})
    def test_get_resource_object_doc_lines(self):
        actual = base.get_resource_object_doc_lines()
        expected = [
            'Resource Objects',
            '----------------',
            '.. _resource-an-object:',
            '',
            'An Object',
            '#########',
            'ex description f',
            '',
            'Attributes',
            '**********',
            '* **str** (*str*) - auth token',
            '',
            'Example',
            '*******',
            '.. code-block:: json',
            '',
            '   {',
            '       "str": "ex str"',
            '   }'
        ]
        assert expected == actual

    @mock.patch.dict('doctor.docs.base.ALL_RESOURCES', {})
    def test_get_resource_object_doc_lines_no_resources(self):
        """
        This test verifies if we have no resources to document, we don't
        attempt to do it anyway which would result in a header w/ no content
        below it.
        """
        assert [] == base.get_resource_object_doc_lines()



from functools import wraps
from inspect import Parameter

import mock

from doctor.resource import ResourceAnnotation, ResourceSchemaAnnotation

from .base import TestCase
from .types import ItemId
from .utils import add_doctor_attrs


def does_nothing(func):
    """An example decorator that does nothing."""
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)
    return wrapper


def plain_logic(a, b=1):
    return (a, b)


def plain_logic_with_kwargs(a, b=1, **kwargs):
    return (a, b, kwargs)


@does_nothing
def decorated_logic(a, b=1):
    return (a, b)


@does_nothing
def decorated_logic_with_kwargs(a, b=1, **kwargs):
    return (a, b, kwargs)


class TestResourceAnnotation(TestCase):

    def test_init(self):
        def logic(user_id: int, item_id: ItemId):
            pass

        logic = add_doctor_attrs(logic)
        annotation = ResourceAnnotation(logic, 'POST')
        assert annotation.logic == logic
        assert annotation.http_method == 'POST'
        assert annotation.title == 'Create'

        # Verify that the annotated_parameters attribute only has parameters
        # from the signature that are doctor types.  It shouldn't include
        # `user_id` since int doesn't extend doctor.types.SuperType
        expected = Parameter('item_id', Parameter.POSITIONAL_OR_KEYWORD,
                             default=Parameter.empty, annotation=ItemId)
        assert {'item_id': expected} == annotation.annotated_parameters

    def test_init_title(self):
        def logic():
            pass

        logic = add_doctor_attrs(logic)
        tests = (
            # (http_method, title, expected)
            ('GET', None, 'Retrieve'),
            ('POST', None, 'Create'),
            ('PUT', None, 'Update'),
            ('DELETE', None, 'Delete'),
            ('PUT', 'Batch', 'Batch'),
        )
        for http_method, title, expected in tests:
            annotation = ResourceAnnotation(logic, http_method, title=title)
            self.assertEqual(expected, annotation.title)


class TestResourceSchemaAnnotation(TestCase):

    def test_init(self):
        s = mock.sentinel
        annotation = ResourceSchemaAnnotation(
            s.logic, 'POST', s.schema, s.request_schema,
            s.response_schema)
        self.assertEqual(annotation.logic, s.logic)
        self.assertEqual(annotation.http_method, 'POST')
        self.assertEqual(annotation.schema, s.schema)
        self.assertEqual(annotation.request_schema, s.request_schema)
        self.assertEqual(annotation.response_schema, s.response_schema)
        self.assertEqual(annotation.title, 'Create')

    def test_init_title(self):
        tests = (
            # (http_method, title, expected)
            ('GET', None, 'Retrieve'),
            ('POST', None, 'Create'),
            ('PUT', None, 'Update'),
            ('DELETE', None, 'Delete'),
            ('PUT', 'Batch', 'Batch'),
        )
        s = mock.sentinel
        for http_method, title, expected in tests:
            annotation = ResourceSchemaAnnotation(
                s.logic, http_method, s.schema, s.request_schema,
                s.response_schema, title=title)
            self.assertEqual(expected, annotation.title)

    def test_get_annotation(self):
        """Tests that get_annotation works for badly decorated functions."""
        def decorator(fn):
            def wrapper():
                fn()
            return wrapper
        mock_logic = mock.Mock()
        mock_logic._schema_annotation = mock.sentinel.logic_annotation
        wrapper = decorator(mock_logic)
        self.assertEqual(ResourceSchemaAnnotation.get_annotation(mock_logic),
                         mock.sentinel.logic_annotation)
        self.assertEqual(ResourceSchemaAnnotation.get_annotation(wrapper),
                         mock.sentinel.logic_annotation)
        wrapper._schema_annotation = mock.sentinel.wrapper_annotation
        self.assertEqual(ResourceSchemaAnnotation.get_annotation(wrapper),
                         mock.sentinel.wrapper_annotation)
        delattr(wrapper, '_schema_annotation')
        delattr(mock_logic, '_schema_annotation')
        self.assertIsNone(ResourceSchemaAnnotation.get_annotation(wrapper))



"""This module can be used to generate Sphinx documentation for an API."""

from __future__ import absolute_import

import json
import pipes
import re
from collections import defaultdict
from inspect import Parameter
from typing import Any, Dict, List

try:
    from docutils import nodes
    from docutils.parsers.rst import Directive
    from docutils.statemachine import ViewList
    from sphinx.errors import SphinxError
    from sphinx.util.nodes import nested_parse_with_titles
    from sphinxcontrib.autohttp.common import http_directive
    from sphinxcontrib.httpdomain import setup as setup_httpdomain
except ImportError:  # pragma: no cover
    raise ImportError('You must install sphinx and sphinxcontrib-httpdomain to '
                      'use the doctor.docs module.')

from doctor.resource import ResourceAnnotation
from doctor.response import Response
from doctor.types import Array, Enum, Object
from doctor.utils import get_description_lines


#: Used to transform a class name into it's various words, splitting on
#: uppercase characters.  So `MyClassName` becomes ['My', 'Class', 'Name']
CAMEL_CASE_RE = re.compile(r'[A-Z][^A-Z]*')

#: Used to get all url parameter names.
URL_PARAMS_RE = re.compile(r'\(([a-zA-Z_]+)\:')

#: These are the HTTP methods that will be documented on handlers.
#:
#: Note that HEAD and OPTIONS aren't included here.
HTTP_METHODS = ('get', 'post', 'put', 'patch', 'delete')

#: Used to map the JSON schema types to more Pythonic types for consistency.
TYPE_MAP = {'array': 'list',
            'boolean': 'bool',
            'integer': 'int',
            'number': 'float',
            'object': 'dict',
            'string': 'str'}

#: Used to list all types that can have multiple items
# Excluded allOf and multipleOf since they require more than one item
OF_TYPES = {'anyOf', 'oneOf'}

# Any line that begins with this text will be transofrmed into a sphinx heading
HEADING_TOKEN = '!!HEADING!!'
HEADING_TOKEN_LENGTH = len(HEADING_TOKEN)

#: Stores a mapping of resource names to their annotated type.
ALL_RESOURCES = {}


def get_example_curl_lines(method: str, url: str, params: dict,
                           headers: dict) -> List[str]:
    """Render a cURL command for the given request.

    :param str method: HTTP request method (e.g. "GET").
    :param str url: HTTP request URL.
    :param dict params: JSON body, for POST and PUT requests.
    :param dict headers: A dict of HTTP headers.
    :returns: list
    """
    parts = ['curl {}'.format(pipes.quote(url))]
    parts.append('-X {}'.format(method))
    for header in headers:
        parts.append("-H '{}: {}'".format(header, headers[header]))
    if method not in ('DELETE', 'GET'):
        # Don't append a json body if there are no params.
        if params:
            parts.append("-H 'Content-Type: application/json' -d")
            pretty_json = json.dumps(params, separators=(',', ': '), indent=4,
                                     sort_keys=True)
            # add indentation for the closing bracket of the json body
            json_lines = pretty_json.split('\n')
            json_lines[-1] = '   ' + json_lines[-1]
            pretty_json = '\n'.join(json_lines)
            parts.append(pipes.quote(pretty_json))
    wrapped = [parts.pop(0)]
    for part in parts:
        if len(wrapped[-1]) + len(part) < 80:
            wrapped[-1] += ' ' + part
        else:
            wrapped[-1] += ' \\'
            wrapped.append('  ' + part)
    return wrapped


def get_example_lines(headers: Dict[str, str], method: str, url: str,
                      params: Dict[str, Any], response: str) -> List[str]:
    """Render a reStructuredText example for the given request and response.

    :param dict headers: A dict of HTTP headers.
    :param str method: HTTP request method (e.g. "GET").
    :param str url: HTTP request URL.
    :param dict params: Form parameters, for POST and PUT requests.
    :param str response: Text response body.
    :returns: list
    """
    lines = ['', 'Example Request:', '', '.. code-block:: bash', '']
    lines.extend(prefix_lines(
        get_example_curl_lines(method, url, params, headers), '   '))
    lines.extend(['', 'Example Response:', ''])
    try:
        # Try to parse and prettify the response as JSON. If it fails
        # (for whatever reason), we'll treat it as text instead.
        response = json.dumps(json.loads(response), indent=2,
                              separators=(',', ': '), sort_keys=True)
        lines.extend(['.. code-block:: json', ''])
    except Exception:
        lines.extend(['.. code-block:: text', ''])
    lines.extend(prefix_lines(response, '   '))
    return lines


def get_json_object_lines(annotation: ResourceAnnotation,
                          properties: Dict[str, Any], field: str,
                          url_params: Dict, request: bool=False,
                          object_property: bool=False) -> List[str]:
    """Generate documentation for the given object annotation.

    :param doctor.resource.ResourceAnnotation annotation:
        Annotation object for the associated handler method.
    :param str field: Sphinx field type to use (e.g. '<json').
    :param list url_params: A list of url parameter strings.
    :param bool request: Whether the schema is for the request or not.
    :param bool object_property: If True it indicates this is a property of
        an object that we are documenting.  This is only set to True when
        called recursively when encountering a property that is an object in
        order to document the properties of it.
    :returns: list of strings, one for each line.
    """
    required_lines = []
    lines = []
    default_field = field
    for prop in sorted(properties.keys()):
        annotated_type = properties[prop]
        # If the property is a url parameter override the field to use
        # param so that it's not documented in the json body or query params.
        field = default_field
        if request and prop in url_params:
            field = 'param'

        types = [str(annotated_type.native_type.__name__)]
        description = annotated_type.description
        obj_ref = ''
        if issubclass(annotated_type, Object):
            resource_name = annotated_type.title
            if resource_name is None:
                class_name = annotated_type.__name__
                resource_name = class_name_to_resource_name(class_name)
            ALL_RESOURCES[resource_name] = annotated_type
            obj_ref = ' See :ref:`resource-{}`.'.format(
                '-'.join(resource_name.split(' ')).lower().strip())
        elif (hasattr(annotated_type, 'items') and
                issubclass(annotated_type.items, Object)):
            # This means the type is an array of objects, so we want to
            # collect the object as a resource we can document later.
            resource_name = annotated_type.items.title
            if resource_name is None:
                class_name = annotated_type.items.__name__
                resource_name = class_name_to_resource_name(class_name)
            ALL_RESOURCES[resource_name] = annotated_type.items
            obj_ref = ' See :ref:`resource-{}`.'.format(
                '-'.join(resource_name.split(' ')).lower().strip())

        # Document any enum.
        enum = ''
        if issubclass(annotated_type, Enum):
            enum = ' Must be one of: `{}`.'.format(annotated_type.enum)

        field_prop = prop
        # If this is a request param and the property is required
        # add required text and append lines to required_lines.  This
        # will make the required properties appear in alphabetical order
        # before the optional.
        if request and prop in annotation.params.required:
            description = '**Required**.  ' + description
            required_lines.append(
                ':{field} {types} {prop}: {description}{enum}{obj_ref}'.format(
                    field=field, types=','.join(types), prop=field_prop,
                    description=description, enum=enum, obj_ref=obj_ref))
        else:
            lines.append(
                ':{field} {types} {prop}: {description}{enum}{obj_ref}'.format(
                    field=field, types=','.join(types), prop=field_prop,
                    description=description, enum=enum, obj_ref=obj_ref))

    return required_lines + lines


def get_json_lines(annotation: ResourceAnnotation, field: str, route: str,
                   request: bool=False) -> List:
    """Generate documentation lines for the given annotation.

    This only documents schemas of type "object", or type "list" where each
    "item" is an object. Other types are ignored (but a warning is logged).

    :param doctor.resource.ResourceAnnotation annotation:
        Annotation object for the associated handler method.
    :param str field: Sphinx field type to use (e.g. '<json').
    :param str route: The route the annotation is attached to.
    :param bool request: Whether the resource annotation is for the request or
        not.
    :returns: list of strings, one for each line.
    """
    url_params = URL_PARAMS_RE.findall(route)
    if not request:
        return_type = annotation.logic._doctor_signature.return_annotation
        # Check if our return annotation is a Response that supplied a
        # type we can use to document.  If so, use that type for api docs.
        # e.g. def logic() -> Response[MyType]
        if issubclass(return_type, Response):
            if return_type.__args__ is not None:
                return_type = return_type.__args__[0]
        if issubclass(return_type, Array):
            if issubclass(return_type.items, Object):
                properties = return_type.items.properties
                field += 'arr'
            else:
                return []
        elif issubclass(return_type, Object):
            properties = return_type.properties
        else:
            return []
    else:
        # If we defined a req_obj_type for the logic, use that type's
        # properties instead of the function signature.
        if annotation.logic._doctor_req_obj_type:
            properties = annotation.logic._doctor_req_obj_type.properties
        else:
            parameters = annotation.annotated_parameters
            properties = {k: p.annotation for k, p in parameters.items()}
    return get_json_object_lines(annotation, properties, field, url_params,
                                 request)


def get_resource_object_doc_lines() -> List[str]:
    """Generate documentation lines for all collected resource objects.

    As API documentation is generated we keep a running list of objects used
    in request parameters and responses.  This section will generate
    documentation for each object and provide an inline reference in the API
    documentation.

    :returns: A list of lines required to generate the documentation.
    """
    # First loop through  all resources and make sure to add any properties that
    # are objects and not already in `ALL_RESOURCES`.  We iterate over a copy
    # since we will be modifying the dict during the loop.
    for resource_name, a_type in ALL_RESOURCES.copy().items():
        for prop_a_type in a_type.properties.values():
            if issubclass(prop_a_type, Object):
                resource_name = prop_a_type.title
                if resource_name is None:
                    class_name = prop_a_type.__name__
                    resource_name = class_name_to_resource_name(class_name)
                ALL_RESOURCES[resource_name] = prop_a_type
            elif (hasattr(prop_a_type, 'items') and
                    issubclass(prop_a_type.items, Object)):
                # This means the type is an array of objects, so we want to
                # collect the object as a resource we can document later.
                resource_name = prop_a_type.items.title
                if resource_name is None:
                    class_name = prop_a_type.items.__name__
                    resource_name = class_name_to_resource_name(class_name)
                ALL_RESOURCES[resource_name] = prop_a_type.items

    # If we don't have any resources to document, just return.
    if not ALL_RESOURCES:
        return []

    lines = ['Resource Objects', '----------------']
    for resource_name in sorted(ALL_RESOURCES.keys()):
        a_type = ALL_RESOURCES[resource_name]
        # First add a reference to the resource
        resource_ref = '_resource-{}'.format(
            '-'.join(resource_name.lower().split(' ')))
        lines.extend(['.. {}:'.format(resource_ref), ''])
        # Add resource name heading
        lines.extend([resource_name, '#' * len(resource_name)])
        # Add resource description
        lines.extend([a_type.description, ''])
        # Only document attributes if it has properties defined.
        if a_type.properties:
            # Add attributes documentation.
            lines.extend(['Attributes', '**********'])
            for prop in a_type.properties:
                prop_a_type = a_type.properties[prop]
                description = a_type.properties[prop].description.strip()
                # Add any object reference if the property is an object or
                # an array of objects.
                obj_ref = ''
                if issubclass(prop_a_type, Object):
                    resource_name = prop_a_type.title
                    if resource_name is None:
                        class_name = prop_a_type.__name__
                        resource_name = class_name_to_resource_name(class_name)
                    obj_ref = ' See :ref:`resource-{}`.'.format(
                        '-'.join(resource_name.split(' ')).lower().strip())
                elif (hasattr(prop_a_type, 'items') and
                        issubclass(prop_a_type.items, Object)):
                    # This means the type is an array of objects, so we want to
                    # collect the object as a resource we can document later.
                    resource_name = prop_a_type.items.title
                    if resource_name is None:
                        class_name = prop_a_type.items.__name__
                        resource_name = class_name_to_resource_name(class_name)
                    obj_ref = ' See :ref:`resource-{}`.'.format(
                        '-'.join(resource_name.split(' ')).lower().strip())
                native_type = a_type.properties[prop].native_type.__name__
                if prop in a_type.required:
                    description = '**Required**.  ' + description
                lines.append('* **{}** (*{}*) - {}{}'.format(
                    prop, native_type, description, obj_ref).strip())
            lines.append('')
        # Add example of object.
        lines.extend(['Example', '*******'])
        example = a_type.get_example()
        pretty_json = json.dumps(example, separators=(',', ': '), indent=4,
                                 sort_keys=True)
        pretty_json_lines = prefix_lines(pretty_json, '   ')
        lines.extend(['.. code-block:: json', ''])
        lines.extend(pretty_json_lines)
    return lines


def get_name(value) -> str:
    """Return a best guess at the qualified name for a class or function.

    :param value: A class or function object.
    :type value: class or function
    :returns str:
    """
    if value.__module__ == '__builtin__':
        return value.__name__
    else:
        return '.'.join((value.__module__, value.__name__))


def normalize_route(route: str) -> str:
    """Strip some of the ugly regexp characters from the given pattern.

    >>> normalize_route('^/user/<user_id:int>/?$')
    u'/user/(user_id:int)/'
    """
    normalized_route = str(route).lstrip('^').rstrip('$').rstrip('?')
    normalized_route = normalized_route.replace('<', '(').replace('>', ')')
    return normalized_route


def prefix_lines(lines, prefix):
    """Add the prefix to each of the lines.

    >>> prefix_lines(['foo', 'bar'], '  ')
    ['  foo', '  bar']
    >>> prefix_lines('foo\\nbar', '  ')
    ['  foo', '  bar']

    :param list or str lines: A string or a list of strings. If a string is
        passed, the string is split using splitlines().
    :param str prefix: Prefix to add to the lines. Usually an indent.
    :returns: list
    """
    if isinstance(lines, bytes):
        lines = lines.decode('utf-8')
    if isinstance(lines, str):
        lines = lines.splitlines()
    return [prefix + line for line in lines]


def class_name_to_resource_name(class_name: str) -> str:
    """Converts a camel case class name to a resource name with spaces.

    >>> class_name_to_resource_name('FooBarObject')
    'Foo Bar Object'

    :param class_name: The name to convert.
    :returns: The resource name.
    """
    s = re.sub('(.)([A-Z][a-z]+)', r'\1 \2', class_name)
    return re.sub('([a-z0-9])([A-Z])', r'\1 \2', s)


class BaseDirective(Directive):

    """Base class for doctor Sphinx directives.

    You probably want to use
    :class:`~doctor.docs.flask.AutoFlaskDirective` instead of
    this class.
    """

    #: Name to use for this directive.
    #:
    #: This is the identifier used within the Sphinx documentation to trigger
    #: this directive. This value should be set by subclasses. For example, in
    #: :class:`~doctor.flask.docs.AutoFlaskDirective`, this is
    #: set to "autoflask".
    directive_name = None

    #: Harness for the Flask app this directive is documenting. This
    #: is responsible for setting up and tearing down the mock app. It is
    #: defined in Sphinx's conf.py file, and set on the directive in
    #: :meth:`~doctor.docs.base.BaseDirective.run_setup`. It should
    #: be an instance of :class:`~doctor.docs.base.BaseHarness`.
    harness = None

    #: Indicates to Sphinx that this directive will yield content.
    has_content = True

    def _make_example(
            self, route, handler, annotation):  # pragma: no cover
        self.harness.setup_request(self, route, handler, annotation)
        headers = self.harness._get_headers(str(route), annotation)
        try:
            response_data = self.harness.request(route, handler, annotation)
            return get_example_lines(headers, **response_data)
        except Exception as e:
            raise SphinxError(
                'Error rendering {method} {route} example: {exc}'.format(
                    method=annotation.http_method, route=route,
                    exc=e)
            )
        finally:
            self.harness.teardown_request(self, route, handler, annotation)

    def _prepare_env(self):  # pragma: no cover
        """Setup the document's environment, if necessary."""
        env = self.state.document.settings.env
        if not hasattr(env, self.directive_name):
            # Track places where we use this directive, so we can check for
            # outdated documents in the future.
            state = DirectiveState()
            setattr(env, self.directive_name, state)
        else:
            state = getattr(env, self.directive_name)
        return env, state

    def _render_rst(self):  # pragma: no cover
        """Render lines of reStructuredText for items yielded by
        :meth:`~doctor.docs.base.BaseHarness.iter_annotations`.
        """
        # Create a mapping of headers to annotations.  We want to group
        # all annotations by a header, but they could be in multiple handlers
        # so we create a map of them here with the heading as the key and
        # the list of associated annotations as a list.  This is so we can
        # sort them alphabetically to make reading the api docs easier.
        heading_to_annotations_map = defaultdict(list)
        for heading, route, handler, annotations in (
                self.harness.iter_annotations()):
            # Set the route and handler as attributes so we can retrieve them
            # when we loop through them all below.
            for annotation in annotations:
                annotation.route = route
                annotation.handler = handler
                heading_to_annotations_map[heading].append(annotation)

        headings = list(heading_to_annotations_map.keys())
        headings.sort()
        previous_heading = None
        for heading in headings:
            annotations = heading_to_annotations_map.get(heading)
            # Sort all the annotations by title.
            annotations.sort(key=lambda a: a.title)
            # Only emit a new heading if the resource has changed.  This
            # esnures that documented endpoints for the same resource all
            # end up under a single heading.
            if previous_heading != heading:
                previous_heading = heading
                yield HEADING_TOKEN + heading
            for annotation in annotations:
                route = annotation.route
                normalized_route = normalize_route(route)
                handler = annotation.handler
                # Adds a title for the endpoint.
                if annotation.title is not None:
                    yield annotation.title
                    yield '#' * len(annotation.title)
                docstring = get_description_lines(getattr(annotation.logic,
                                                          '__doc__', None))

                # Documents the logic function associated with the annotation.
                docstring.append(':Logic Func: :func:`~{}.{}`'.format(
                    annotation.logic.__module__, annotation.logic.__name__))

                field = '<json'
                if annotation.http_method in ('DELETE', 'GET'):
                    field = 'query'
                docstring.extend(get_json_lines(
                    annotation, field=field, route=normalized_route,
                    request=True)
                )

                # Document any request headers.
                defined_headers = list(self.harness._get_headers(
                    str(route), annotation).keys())
                defined_headers.sort()
                for header in defined_headers:
                    definition = self.harness.header_definitions.get(
                        header, '').strip()
                    docstring.append(':reqheader {}: {}'.format(
                        header, definition))

                # Document response if a type was defined.
                if annotation.return_annotation != Parameter.empty:
                    docstring.extend(get_json_lines(
                        annotation, field='>json', route=normalized_route))

                docstring.extend(self._make_example(route, handler, annotation))
                for line in http_directive(annotation.http_method,
                                           normalized_route, docstring):
                    yield line

        # Document resource objects.
        for line in get_resource_object_doc_lines():
            yield line

    def run(self):  # pragma: no cover
        """Called by Sphinx to generate documentation for this directive."""
        if self.directive_name is None:
            raise NotImplementedError('directive_name must be implemented by '
                                      'subclasses of BaseDirective')
        env, state = self._prepare_env()
        state.doc_names.add(env.docname)
        directive_name = '<{}>'.format(self.directive_name)
        node = nodes.section()
        node.document = self.state.document
        result = ViewList()
        for line in self._render_rst():
            if line.startswith(HEADING_TOKEN):
                # Remove heading token, then append 2 lines, one with
                # the heading text, and the other with the dashes to
                # underline the heading.
                heading = line[HEADING_TOKEN_LENGTH:]
                result.append(heading, directive_name)
                result.append('-' * len(heading), directive_name)
            else:
                result.append(line, directive_name)
        nested_parse_with_titles(self.state, result, node)
        return node.children

    @classmethod
    def get_outdated_docs(
            cls, app, env, added, changed, removed):  # pragma: no cover
        """Handler for Sphinx's env-get-outdated event.

        This handler gives a Sphinx extension a chance to indicate that some
        set of documents are out of date and need to be re-rendered. The
        implementation here is stupid, for now, and always says that anything
        that uses the directive needs to be re-rendered.

        We should make it smarter, at some point, and have it figure out which
        modules are used by the associated handlers, and whether they have
        actually been updated since the last time the given document was
        rendered.
        """
        state = getattr(env, cls.directive_name, None)
        if state and state.doc_names:
            # This is stupid for now, and always says everything that uses
            # this autodoc generation needs to be updated. We should make this
            # smarter at some point and actually figure out what modules are
            # touched, and whether they have been changed.
            return sorted(state.doc_names)
        else:
            return []

    @classmethod
    def purge_docs(cls, app, env, docname):  # pragma: no cover
        """Handler for Sphinx's env-purge-doc event.

        This event is emitted when all traces of a source file should be cleaned
        from the environment (that is, if the source file is removed, or before
        it is freshly read). This is for extensions that keep their own caches
        in attributes of the environment.

        For example, there is a cache of all modules on the environment. When a
        source file has been changed, the cache's entries for the file are
        cleared, since the module declarations could have been removed from the
        file.
        """
        state = getattr(env, cls.directive_name, None)
        if state and docname in state.doc_names:
            state.doc_names.remove(docname)

    @classmethod
    def run_setup(cls, app):  # pragma: no cover
        config_attr = '{}_harness'.format(cls.directive_name)
        cls.harness = getattr(app.config, config_attr)
        if not cls.harness:
            raise SphinxError('Please set {} in your conf.py'.format(
                config_attr))
        cls.harness.setup_app(app)

    @classmethod
    def run_teardown(cls, app, exception):  # pragma: no cover
        cls.harness.teardown_app(app)

    @classmethod
    def setup(cls, app):  # pragma: no cover
        """Called by Sphinx to setup an extension."""
        if cls.directive_name is None:
            raise NotImplementedError('directive_name must be set by '
                                      'subclasses of BaseDirective')
        if not app.registry.has_domain('http'):
            setup_httpdomain(app)
        app.add_config_value('{}_harness'.format(cls.directive_name),
                             None, 'env')
        app.add_directive(cls.directive_name, cls)
        app.connect('builder-inited', cls.run_setup)
        app.connect('build-finished', cls.run_teardown)
        app.connect('env-get-outdated', cls.get_outdated_docs)
        app.connect('env-purge-doc', cls.purge_docs)


class BaseHarness(object):

    """
    Base class for doctor directive harnesses. A harness is defined
    in Sphinx's conf.py, and the directive invokes the various methods at the
    appropriate times, so the app can bootstrap a mock version of itself.
    """

    def __init__(self, url_prefix):
        super(BaseHarness, self).__init__()
        self.defined_example_values = {}
        self.url_prefix = url_prefix.rstrip('/')
        #: Stores headers for particular methods and routes.
        self.defined_header_values = {}
        #: Stores global headers to use on all requests
        self.headers = {}
        #: Stores definitions for header keys for documentation.
        self.header_definitions = {}

    def iter_annotations(self):  # pragma: no cover
        """Yield a tuple for each schema annotated handler to document.

        This must be implemented by subclasses. See
        :class:`~doctor.docs.flask.AutoFlaskHarness` for an
        example implementation.
        """
        raise NotImplementedError('This method must be implemented by '
                                  'subclasses of BaseHarness')

    def define_header_values(self, http_method, route, values, update=False):
        """Define header values for a given request.

        By default, header values are determined from the class attribute
        `headers`. But if you want to change the headers used in the
        documentation for a specific route, this method lets you do that.

        :param str http_method: An HTTP method, like "get".
        :param str route: The route to match.
        :param dict values: A dictionary of headers for the example request.
        :param bool update: If True, the values will be merged into the default
            headers for the request. If False, the values will replace
            the default headers.
        """
        self.defined_header_values[(http_method.lower(), route)] = {
            'update': update,
            'values': values
        }

    def define_example_values(self, http_method, route, values, update=False):
        """Define example values for a given request.

        By default, example values are determined from the example properties
        in the schema. But if you want to change the example used in the
        documentation for a specific route, and this method lets you do that.

        :param str http_method: An HTTP method, like "get".
        :param str route: The route to match.
        :param dict values: A dictionary of parameters for the example request.
        :param bool update: If True, the values will be merged into the default
            example values for the request. If False, the values will replace
            the default example values.
        """
        self.defined_example_values[(http_method.lower(), route)] = {
            'update': update,
            'values': values
        }

    def request(self, route, handler, annotation):
        """Make a request against the app.

        This must be implemented by subclasses. See
        :class:`~doctor.docs.flask.AutoFlaskHarness` for an
        example implementation.
        """
        raise NotImplementedError('This method must be implemented by '
                                  'subclasses of BaseHarness')

    def setup_app(self, sphinx_app):  # pragma: no cover
        """Called once before building documentation.

        :param sphinx_app: Sphinx application object.
        """
        pass

    def setup_request(self, sphinx_directive, route, handler,
                      annotation):   # pragma: no cover
        """Called before each request to the mock app.

        :param BaseDirective sphinx_directive: The directive that is making
            the mock request.
        :param route: Path for the route. For Flask, this will be a Route
            object.
        :param handler: Flask resource for the route.
        :param ResourceAnnotation annotation: Annotation for the request.
        """
        pass

    def teardown_app(self, sphinx_app):  # pragma: no cover
        """Called once after building documentation.

        :param sphinx_app: Sphinx application object.
        """
        pass

    def teardown_request(self, sphinx_directive, route, handler,
                         annotation):   # pragma: no cover
        """Called after each request to the mock app.

        :param BaseDirective sphinx_directive: The directive that is making
            the mock request.
        :param route: Path for the route. For Flask, this will be a Route
            object.
        :param handler: Flask resource for the route.
        :param ResourceAnnotation annotation: Annotation for the request.
        """
        pass

    def _get_annotation_heading(self, handler, route, heading=None):
        """Returns the heading text for an annotation.

        Attempts to get the name of the heading from the handler attribute
        `schematic_title` first.

        If `schematic_title` it is not present, it attempts to generate
        the title from the class path.
        This path: advertiser_api.handlers.foo_bar.FooListHandler
        would translate to 'Foo Bar'

        If the file name with the resource is generically named handlers.py
        or it doesn't have a full path then we attempt to get the resource
        name from the class name.
        So FooListHandler and FooHandler would translate to 'Foo'.
        If the handler class name starts with 'Internal', then that will
        be appended to the heading.
        So InternalFooListHandler would translate to 'Foo (Internal)'

        :param mixed handler: The handler class.  Will be a flask resource class
        :param str route: The route to the handler.
        :returns: The text for the heading as a string.
        """
        if hasattr(handler, '_doctor_heading'):
            return handler._doctor_heading

        heading = ''
        handler_path = str(handler)
        try:
            handler_file_name = handler_path.split('.')[-2]
        except IndexError:
            # In the event there is no path and we just have the class name,
            # get heading from the class name by setting us up to enter the
            # first if statement.
            handler_file_name = 'handler'

        # Get heading from class name
        if handler_file_name.startswith('handler'):
            class_name = handler_path.split('.')[-1]
            internal = False
            for word in CAMEL_CASE_RE.findall(class_name):
                if word == 'Internal':
                    internal = True
                    continue
                elif word.startswith(('List', 'Handler', 'Resource')):
                    # We've hit the end of the class name that contains
                    # words we are interested in.
                    break
                heading += '%s ' % (word,)
            if internal:
                heading = heading.strip()
                heading += ' (Internal)'
        # Get heading from handler file name
        else:
            heading = ' '.join(handler_file_name.split('_')).title()
            if 'internal' in route:
                heading += ' (Internal)'
        return heading.strip()

    def _get_headers(self, route: str, annotation: ResourceAnnotation) -> Dict:
        """Gets headers for the provided route.

        :param route: The route to get example values for.
        :type route: werkzeug.routing.Rule for a flask api.
        :param annotation: Schema annotation for the method to be requested.
        :type annotation: doctor.resource.ResourceAnnotation
        :retruns: A dict containing headers.
        """
        headers = self.headers.copy()
        defined_header_values = self.defined_header_values.get(
            (annotation.http_method.lower(), str(route)))
        if defined_header_values is not None:
            if defined_header_values['update']:
                headers.update(defined_header_values['values'])
            else:
                headers = defined_header_values['values']
        return headers

    def _get_example_values(self, route: str,
                            annotation: ResourceAnnotation) -> Dict[str, Any]:
        """Gets example values for all properties in the annotation's schema.

        :param route: The route to get example values for.
        :type route: werkzeug.routing.Rule for a flask api.
        :param annotation: Schema annotation for the method to be requested.
        :type annotation: doctor.resource.ResourceAnnotation
        :retruns: A dict containing property names as keys and example values
            as values.
        """
        defined_values = self.defined_example_values.get(
            (annotation.http_method.lower(), str(route)))
        if defined_values and not defined_values['update']:
            return defined_values['values']

        # If we defined a req_obj_type for the logic, use that type's
        # example values instead of the annotated parameters.
        if annotation.logic._doctor_req_obj_type:
            values = annotation.logic._doctor_req_obj_type.get_example()
        else:
            values = {
                k: v.annotation.get_example()
                for k, v in annotation.annotated_parameters.items()
            }
        if defined_values:
            values.update(defined_values['values'])

        # If this is a GET route, we need to json dumps any parameters that
        # are lists or dicts.  Otherwise we'll get a 400 error for those params
        if annotation.http_method == 'GET':
            for k, v in values.items():
                if isinstance(v, (list, dict)):
                    values[k] = json.dumps(v)
        return values


class DirectiveState(object):

    """This is used to hold Sphinx serialized state for our directives."""

    def __init__(self):  # pragma: no cover
        super(DirectiveState, self).__init__()
        self.doc_names = set()
